/*!
 * \file
 * \brief p2p benchmark
 * \details Benchmarks p2p communications between all pairs 
 * 
 * \b p2p.plot draws the graph of the execution time (sec) against message size (kb).
 * - input: p2p.out
 * - output: 0-1.eps (0-1 with error bars), 0-1-2.eps (comparison of 0-1, 0-2, 1-2)
 */

#include "mpib.h"
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

int main(int argc, char** argv)
{
	MPI_Init(&argc, &argv);

	int rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

	int exit = 0;
	int stride = 1024;
	int max_size = 102400;
	int parallel = 1;
	MPIB_precision precision = {5, 100, 0.95, 0.025};

	if (rank == 0)
	{
		int c;
		while ((c = getopt(argc, argv, "hs:m:pr:R:c:e:")) != -1)
		{
			switch (c)
			{
				case 'h':
					fprintf
					(
						stderr,
"usage: mpirun [mpirun options] %s [options]\n\
-h			help\n\
-s I			message size stride (default: %d)\n\
-m I			maximum message size (default: %d)\n\
-p			parallel p2p benchmarking (default: %d)\n\
-r I			minimum number of repetitions (default: %d)\n\
-R I			maximum number of repetitions (default: %d)\n\
-c D			confidence level: 0 < D < 1 (default: %le)\n\
-e D			error: 0 < D < 1 (default: %le)\n",
						argv[0],
						stride,
						max_size,
						parallel,
						precision.min_reps,
						precision.max_reps,
						precision.cl,
						precision.eps
					);
					exit = 1;
					break;
				case 's':
					stride = atoi(optarg);
					break;
				case 'm':
					max_size = atoi(optarg);
					break;
				case 'p':
					parallel = 1;
					break;
				case 'r':
					precision.min_reps = atoi(optarg);
					if (precision.max_reps < precision.min_reps)
						precision.max_reps = precision.min_reps;
					break;
				case 'R':
					precision.max_reps = atoi(optarg);
					if (precision.min_reps > precision.max_reps)
						precision.min_reps = precision.max_reps;
					break;
				case 'c':
					precision.cl = atof(optarg);
					break;
				case 'e':
					precision.eps = atof(optarg);
					break;
			}
		}
	}
	MPI_Bcast(&exit, 1, MPI_INT, 0, MPI_COMM_WORLD);
	if (exit)
		goto exit;

	int size;
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	if (size < 2)
	{
		if (rank == 0)
			fprintf(stderr, "Cannot perform p2p benchmarks for %d processes (must be >= 2)\n", size);
		goto exit;
	}

	MPI_Bcast(&stride, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&max_size, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&parallel, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&precision, sizeof(MPIB_precision), MPI_CHAR, 0, MPI_COMM_WORLD);

	MPIB_p2p_container* container = MPIB_Send_Recv_container_alloc();
	MPIB_result* results = (MPIB_result*)malloc(sizeof(MPIB_result) * MPIB_C2(size));

	if (rank == 0)
	{
		printf("#%s p2p benchmark\n", parallel ? "parallel" : "sequential");
		MPIB_print_p2p_th(precision, size);
	}
	double time = 0;
	int M;
	for (M = 0; M < max_size; M += stride)
	{
		int start = MPI_Wtime();
		MPIB_measure_p2p(container, MPI_COMM_WORLD, M, parallel, precision, results);
		time += MPI_Wtime() - start;
		if (rank == 0)
			MPIB_print_p2p_tr(M, size, results);
	}
	if (rank == 0)
	{
		printf("#total benchmarking time: %le\n", time);
		printf("\n\n");
	}

	M = max_size / 2; 
	MPIB_measure_p2p(container, MPI_COMM_WORLD, M, parallel, precision, results);
	if (rank == 0)
	{
		printf("#per p2p\n");
		MPIB_print_p2p_table(M, precision, size, results);
	}

	free(results);
	container->free(container);

exit:
	MPI_Finalize();
	return 0;
}
