#ifndef MPIB_COLLECTIVE_H_
#define MPIB_COLLECTIVE_H_

/*!
 * \defgroup collective Implementations and tests for MPI collective operations
 * \{
 */

/*! \brief Scatter */
typedef int (*MPIB_Scatter)(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype, 
	int root, MPI_Comm comm);

/*! \brief Gather */
typedef int (*MPIB_Gather)(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm);

/*!
 * \brief Checks a scatter implementation.
 * \param scatter scatter implementation
 * \param comm communicator
 * \param root root process
 * \param res result (0 - ok)
 */
void MPIB_test_scatter(MPIB_Scatter scatter, MPI_Comm comm, int root, int* res);

/*!
 * \brief Checks a gather implementation.
 * \param gather gather implementation
 * \param comm communicator
 * \param root root process
 * \param res result (0 - ok)
 */
void MPIB_test_gather(MPIB_Gather gather, MPI_Comm comm, int root, int* res);

/*! \brief Linear MPI_Scatter */
int MPIB_Scatter_linear(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype, 
	int root, MPI_Comm comm);

/*! \brief Linear MPI_Gather */
int MPIB_Gather_linear(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm);

/*!
 * \brief Builds a segment of a tree for the process rank.
 * \details The processes' ranks are arranged in ascending order, starting with 0.
 * If root is not equal to 0, the root and 0 processes are interchanged in the tree.
 * \param size communicatot size
 * \param root root process
 * \param rank current process
 * \param block message size to receive from parent process (size at root)
 * \param parent parent process (-1 at root)
 * \param count number of processes to send messages
 * \param children array of processes to send messages 
 * \param blocks array of message sizes to send
 */
typedef void (*MPIB_tree)(int size, int root, int rank, int* block, int* parent, int* count, int** children, int** blocks);

/*! \brief Tree-based MPI_Scatter */
int MPIB_Scatter_tree(MPIB_tree tree, void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype, 
	int root, MPI_Comm comm);

/*! \brief Tree-based MPI_Gather */
int MPIB_Gather_tree(MPIB_tree tree, void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm);

/*!
 * \brief Binomial MPI_Scatter
 * \details
 * \dontinclude src/mpib_collective_tree.c
 * \skip return MPIB_Scatter_tree
 * \until );
 */
int MPIB_Scatter_binomial(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype, 
	int root, MPI_Comm comm);

/*!
 * \brief Binomial MPI_Gather
 * \details
 * \dontinclude src/mpib_collective_tree.c
 * \skip return MPIB_Gather_tree
 * \until );
 */
int MPIB_Gather_binomial(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm);

/*! \brief Bcast */
typedef int (*MPIB_Bcast)(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm);

/*!
 * \}
 */

#endif /*MPIB_COLLECTIVE_H_*/
