#include "mpib.h"
#include <memory.h>
#include <malloc.h>

/*! \brief Swaps two memory blocks of size n. */
void MPIB_memswap(void* ptr1, void* ptr2, size_t n)
{
	void* tmp = malloc(n);
	memcpy(tmp, ptr1, n);
	memcpy(ptr1, ptr2, n);
	memcpy(ptr2, tmp, n);
	free(tmp);
}

int MPIB_Scatter_tree(MPIB_tree tree, void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype, 
	int root, MPI_Comm comm)
{
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);

	int block;
	int parent;
	int count;
	int* children;
	int* blocks;
	tree(size, root, rank, &block, &parent, &count, &children, &blocks);

	MPI_Aint extent;
	MPI_Type_extent(rank == root ? sendtype : recvtype, &extent);
	int inc = rank == root ? sendcount * extent : recvcount * extent;
	int err = MPI_SUCCESS;
	if (rank == root)
	{
		if (root != 0)
			MPIB_memswap(sendbuf, sendbuf + inc * root, inc);
		{
			MPI_Aint extent;
			MPI_Type_extent(recvtype, &extent);
			int inc = recvcount * extent;
			memcpy(recvbuf, sendbuf, inc);
		}
		int i;
		int pos;
		for (i = count - 1, pos = block - blocks[i]; i >= 0; pos -= blocks[--i])
			err &= MPIB_Send(sendbuf + inc * pos, sendcount * blocks[i], sendtype, children[i], 0, comm);
		if (root != 0)
			MPIB_memswap(sendbuf, sendbuf + inc * root, inc);
	}
	else
	{
		void* ptr = malloc(inc * block);
		err &= MPIB_Recv(ptr, recvcount * block, recvtype, parent, 0, comm, MPI_STATUS_IGNORE);
		memcpy(recvbuf, ptr, inc);
		int i;
		int pos;
		for (i = count - 1, pos = block - blocks[i]; i >= 0; pos -= blocks[--i])
			err &= MPIB_Send(ptr + inc * pos, recvcount * blocks[i], recvtype, children[i], 0, comm);
		free(ptr);
	}

	free(children);
	free(blocks);

	return err;
}

int MPIB_Gather_tree(MPIB_tree tree, void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);

	int block;
	int parent;
	int count;
	int* children;
	int* blocks;
	tree(size, root, rank, &block, &parent, &count, &children, &blocks);

	MPI_Aint extent;
	MPI_Type_extent(rank == root ? recvtype : sendtype, &extent);
	int inc = rank == root ? recvcount * extent : sendcount * extent;
	int err = MPI_SUCCESS;
	if (rank == root)
	{
		{
			MPI_Aint extent;
			MPI_Type_extent(sendtype, &extent);
			int inc = sendcount * extent;
			memcpy(recvbuf, sendbuf, inc);
		}
		int i;
		int pos;
		for (i = count - 1, pos = size - blocks[i]; i >= 0; pos -= blocks[--i])
			err &= MPIB_Recv(recvbuf + inc * pos, recvcount * blocks[i], recvtype, children[i], 0, comm, MPI_STATUS_IGNORE);
		if (root != 0)
			MPIB_memswap(recvbuf, recvbuf + inc * root, inc);
	}
	else
	{
		void* ptr = malloc(inc * block);
		memcpy(ptr, sendbuf, inc);
		int i;
		int pos;
		for (i = count - 1, pos = block - blocks[i]; i >= 0; pos -= blocks[--i])
			err &= MPIB_Recv(ptr + inc * pos, sendcount * block, sendtype, children[i], 0, comm, MPI_STATUS_IGNORE);
		err &= MPIB_Send(ptr, sendcount * block, sendtype, parent, 0, comm);
		free(ptr);
	}
	
	free(children);
	free(blocks);

	return err;
}

/*! \brief Builds a segment of a binomial tree for the process rank. */
void MPIB_binomial_tree(int size, int root, int rank, int* block, int* parent, int* count, int** children, int** blocks)
{
	int vrank = rank;
	if (root != 0)
	{
		if (rank == root)
			vrank = 0;
		if (rank == 0)
			vrank = root;
	}

	int b = size;
	int p = -1;
	int c = 0;
	int* cs = NULL;
	int* bs = NULL;

	int index = 0;
	do
	{
		int i;
		for (i = c - 1; i >= 0; i--)
		{
			if (vrank >= cs[i])
			{
				p = index;
				index = cs[i];
				b = bs[i];
				c = 0;
				break;
			}
		}
		int tmp, bin, pos;
		for (tmp = b - 1, bin = 1, pos = 1; tmp > 0; tmp /= 2, bin *= 2)
		{
			if (tmp % 2)
			{
				cs = realloc(cs, sizeof(int) * (c + 1));
				cs[c] = pos + index;
				bs = realloc(bs, sizeof(int) * (c + 1));
				bs[c] = bin;
				c++;
				pos += bin;
			}
		}
	}
	while (index != vrank);

	if (root != 0)
	{
		if (p == root)
			p = 0;
		else
			if (p == 0)
				p = root;
		int i;
		for (i = 0; i < c; i++)
		{
			if (cs[i] == root)
			{
				cs[i] = 0;
				break;
			}
		}
	} 

	*block = b;
	*parent = p;
	*count = c;
	*children = cs;
	*blocks = bs;
} 

int MPIB_Scatter_binomial(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype, 
	int root, MPI_Comm comm)
{
	return MPIB_Scatter_tree(MPIB_binomial_tree, sendbuf, sendcount, sendtype,
		recvbuf, recvcount, recvtype, root, comm);
}

int MPIB_Gather_binomial(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	return MPIB_Gather_tree(MPIB_binomial_tree, sendbuf, sendcount, sendtype,
		recvbuf, recvcount, recvtype, root, comm);
}
