#include "mpib.h"
#include <malloc.h>
#include <string.h>

void MPIB_container_free(void* this)
{
	free(this);
}

typedef struct MPIB_Send_Recv_container
{
	MPIB_p2p_container base;
	char* buffer;
}
MPIB_Send_Recv_container;

void MPIB_Send_Recv_initialize(void* this, MPI_Comm comm, int M)
{
	MPIB_Send_Recv_container* container = (MPIB_Send_Recv_container*)this;
	container->buffer = (char*)malloc(M);
}

void MPIB_Send_Recv_execute_measure(void* this, MPI_Comm comm, int M, int mirror)
{
	MPIB_Send_Recv_container* container = (MPIB_Send_Recv_container*)this;
	MPIB_Send(container->buffer, M, MPI_CHAR, mirror, 0, comm);
	MPIB_Recv(container->buffer, M, MPI_CHAR, mirror, 0, comm, MPI_STATUS_IGNORE);
}

void MPIB_Send_Recv_execute_mirror(void* this, MPI_Comm comm, int M, int measure)
{
	MPIB_Send_Recv_container* container = (MPIB_Send_Recv_container*)this;
	MPIB_Recv(container->buffer, M, MPI_CHAR, measure, 0, comm, MPI_STATUS_IGNORE);
	MPIB_Send(container->buffer, M, MPI_CHAR, measure, 0, comm);
}

void MPIB_Send_Recv_finalize(void* this, MPI_Comm comm)
{
	MPIB_Send_Recv_container* container = (MPIB_Send_Recv_container*)this;
	free(container->buffer);
}

MPIB_p2p_container* MPIB_Send_Recv_container_alloc()
{
	MPIB_p2p_container* container =
		(MPIB_p2p_container*)malloc(sizeof(MPIB_Send_Recv_container));
	container->initialize = MPIB_Send_Recv_initialize;
	container->execute_measure = MPIB_Send_Recv_execute_measure;
	container->execute_mirror = MPIB_Send_Recv_execute_mirror;
	container->finalize = MPIB_Send_Recv_finalize;
	container->free = MPIB_container_free;
	return container;
}

typedef struct MPIB_Scatter_Gather_container
{
	MPIB_coll_container base;
	char* buffer;
}
MPIB_Scatter_Gather_container;

void MPIB_Scatter_Gather_initialize(void* this, MPI_Comm comm, int root, int M)
{
	MPIB_Scatter_Gather_container* container = (MPIB_Scatter_Gather_container*)this;
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	container->buffer = rank == root ? (char*)malloc(M * size) : (char*)malloc(M);
}

void MPIB_Scatter_Gather_finalize(void* this, MPI_Comm comm, int root)
{
	MPIB_Scatter_Gather_container* container = (MPIB_Scatter_Gather_container*)this;
	free(container->buffer);
}

typedef struct MPIB_Scatter_container
{
	MPIB_Scatter_Gather_container base;
	MPIB_Scatter scatter;
}
MPIB_Scatter_container;

void MPIB_Scatter_execute(void* this, MPI_Comm comm, int root, int M)
{
	MPIB_Scatter_container* container = (MPIB_Scatter_container*)this;
	container->scatter(container->base.buffer, M, MPI_CHAR,
		container->base.buffer, M, MPI_CHAR,
		root, comm); 
}

typedef struct MPIB_Gather_container
{
	MPIB_Scatter_Gather_container base;
	MPIB_Gather gather;
}
MPIB_Gather_container;

void MPIB_Gather_execute(void* this, MPI_Comm comm, int root, int M)
{
	MPIB_Gather_container* container = (MPIB_Gather_container*)this;
	container->gather(container->base.buffer, M, MPI_CHAR,
		container->base.buffer, M, MPI_CHAR,
		root, comm); 
}

void MPIB_Scatter_Gather_init(MPIB_Scatter_Gather_container* container)
{
	container->base.initialize = MPIB_Scatter_Gather_initialize;
	container->base.finalize = MPIB_Scatter_Gather_finalize;
	container->base.free = MPIB_container_free;
}

MPIB_coll_container* MPIB_Scatter_container_alloc(MPIB_Scatter scatter)
{
	MPIB_Scatter_container* container = (MPIB_Scatter_container*)malloc(sizeof(MPIB_Scatter_container));
	MPIB_Scatter_Gather_init((MPIB_Scatter_Gather_container*)container);
	container->base.base.execute = MPIB_Scatter_execute;
	container->scatter = scatter;
	return (MPIB_coll_container*)container;
}

MPIB_coll_container* MPIB_Gather_container_alloc(MPIB_Gather gather)
{
	MPIB_Gather_container* container = (MPIB_Gather_container*)malloc(sizeof(MPIB_Gather_container));
	MPIB_Scatter_Gather_init((MPIB_Scatter_Gather_container*)container);
	container->base.base.execute = MPIB_Gather_execute;
	container->gather = gather;
	return (MPIB_coll_container*)container;
}

typedef struct MPIB_Scatterv_Gatherv_container
{
	MPIB_Scatter_Gather_container base;
	int stride;
	int* counts;
	int* displs;
}
MPIB_Scatterv_Gatherv_container;

void MPIB_Scatterv_Gatherv_initialize(void* this, MPI_Comm comm, int root, int M)
{
	MPIB_Scatterv_Gatherv_container* container = (MPIB_Scatterv_Gatherv_container*)this;
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	container->base.buffer = rank == 0 ?
		(char*)malloc(M * size + container->stride * MPIB_C2(size)) :
		(char*)malloc(M + container->stride * rank);
	container->counts = (int*)malloc(sizeof(int) * size);
	container->displs = (int*)malloc(sizeof(int) * size);
	int i;
	for (i = 0; i < size; i++)
	{
		container->counts[i] = M + container->stride * i;
		container->displs[i] = M * i + container->stride * MPIB_C2(i);
	}
}

void MPIB_Scatterv_Gatherv_finalize(void* this, MPI_Comm comm, int root)
{
	MPIB_Scatterv_Gatherv_container* container = (MPIB_Scatterv_Gatherv_container*)this;
	free(container->base.buffer);
	free(container->counts);
	free(container->displs);
}

void MPIB_Scatterv_execute(void* this, MPI_Comm comm, int root, int M)
{
	MPIB_Scatterv_Gatherv_container* container = (MPIB_Scatterv_Gatherv_container*)this;
	MPI_Scatterv(container->base.buffer, container->counts, container->displs, MPI_CHAR,
		container->base.buffer, M, MPI_CHAR,
		root, comm);
}

void MPIB_Gatherv_execute(void* this, MPI_Comm comm, int root, int M)
{
	MPIB_Scatterv_Gatherv_container* container = (MPIB_Scatterv_Gatherv_container*)this;
	MPI_Gatherv(container->base.buffer, M, MPI_CHAR,
		container->base.buffer, container->counts, container->displs, MPI_CHAR,
		root, comm);
}

void MPIB_Scatterv_Gatherv_init(MPIB_Scatterv_Gatherv_container* container, int stride)
{
	container->base.base.initialize = MPIB_Scatterv_Gatherv_initialize;
	container->base.base.finalize = MPIB_Scatterv_Gatherv_finalize;
	container->base.base.free = MPIB_container_free;
	container->stride = stride;
}

MPIB_coll_container* MPIB_Scatterv_container_alloc(int stride)
{
	MPIB_Scatterv_Gatherv_container* container =
		(MPIB_Scatterv_Gatherv_container*)malloc(sizeof(MPIB_Scatterv_Gatherv_container));
	MPIB_Scatterv_Gatherv_init(container, stride);
	container->base.base.execute = MPIB_Scatterv_execute;
	return (MPIB_coll_container*)container;
}

MPIB_coll_container* MPIB_Gatherv_container_alloc(int stride)
{
	MPIB_Scatterv_Gatherv_container* container =
		(MPIB_Scatterv_Gatherv_container*)malloc(sizeof(MPIB_Scatterv_Gatherv_container));
	MPIB_Scatterv_Gatherv_init(container, stride);
	container->base.base.execute = MPIB_Gatherv_execute;
	return (MPIB_coll_container*)container;
}

MPIB_coll_container* MPIB_coll_container_alloc(const char* operation)
{
	int stride = 256;
	if (!strcmp(operation, "MPI_Scatter"))
		return MPIB_Scatter_container_alloc(MPI_Scatter);
	if (!strcmp(operation, "MPIB_Scatter_linear"))
		return MPIB_Scatter_container_alloc(MPIB_Scatter_linear);
	if (!strcmp(operation, "MPIB_Scatter_binomial"))
		return MPIB_Scatter_container_alloc(MPIB_Scatter_binomial);
	if (!strcmp(operation, "MPI_Gather"))
		return MPIB_Gather_container_alloc(MPI_Gather);
	if (!strcmp(operation, "MPIB_Gather_linear"))
		return MPIB_Gather_container_alloc(MPIB_Gather_linear);
	if (!strcmp(operation, "MPIB_Gather_binomial"))
		return MPIB_Gather_container_alloc(MPIB_Gather_binomial);
	if (!strcmp(operation, "MPI_Scatterv"))
		return MPIB_Scatterv_container_alloc(stride);
	if (!strcmp(operation, "MPI_Gatherv"))
		return MPIB_Gatherv_container_alloc(stride);
	return NULL;
}
