#ifndef MPIB_CONTAINER_H_
#define MPIB_CONTAINER_H_

/*!
 * \defgroup container Containers for the communication operations to be measured
 * \brief Base data structures and some implemented containers 
 * \details How to use:
 * - Create the instance of the container.   
 * - Pass the instance of the container to the measurement function.   
 * - Destroy the instance of the container.   
 * \code
MPIB_p2p_container* container = MPIB_x_container_alloc();
MPIB_measure_p2p(container, comm, M, parallel, precision, result);
container->free(container);
 * \endcode
 * 
 * How to extend:
 * \dontinclude src/mpib_container.c
 * - Create a data structure with the first field \ref MPIB_p2p_container or \ref MPIB_coll_container.
 * \skip typedef struct MPIB_Send_Recv_container
 * \until MPIB_Send_Recv_container;
 * - Implement the functions: initialize, execute*, finalize,
 * where \c this argument can be typecasted to the data structure.
 * \skip MPIB_Send_Recv_initialize
 * \until free(container->buffer);
 * \line }
 * - Implement the functions that allocate and free the data structure.
 * \dontinclude src/mpib_container.c
 * \skip MPIB_container_free
 * \until free(this);
 * \line }
 * \skip MPIB_Send_Recv_container_alloc
 * \until return container;
 * \line }
 * 
 * Container allocators can have arguments, for example \ref MPIB_Scatter_container_alloc has an argument
 * \c MPIB_Scatter, pointer to a scatter implementation. This provides two-level extension of measurement:
 * \dot
	digraph container
	{
		node [fontsize=8];
		MPI_Scatter -> MPIB_Scatter_container;
		MPIB_Scatter_linear -> MPIB_Scatter_container;
		MPIB_Scatter_binomial -> MPIB_Scatter_container;
		MPI_Scatterv -> MPIB_Scatterv_container;
		MPIB_Scatter_container -> MPIB_measure_max;
		MPIB_Scatter_container -> MPIB_measure_root;
		MPIB_Scatter_container -> MPIB_measure_global;
		MPIB_Scatterv_container -> MPIB_measure_max;
		MPIB_Scatterv_container -> MPIB_measure_root;
		MPIB_Scatterv_container -> MPIB_measure_global;
	}
 * \enddot
 * \{
 */ 

/*! \brief
 * Container for a point-to-point communication operation to be measured by \ref MPIB_measure_p2p
 * \param this pointer to itself
 * \param comm MPI communicator over which the communication operation will be performed
 * \param M message size
 * \param mirror mirror process
 * \param measure measure process
 */
typedef struct MPIB_p2p_container
{
	/*!
	 * \brief Initializion of data required for the communication operation
	 * \details Used by \ref MPIB_measure_p2p
	 */
	void (*initialize)(void* this, MPI_Comm comm, int M);
	/*!
	 * \brief Part of communication at the measure side
	 * \details Used by \ref MPIB_measure_p2p
	 */
	void (*execute_measure)(void* this, MPI_Comm comm, int M, int mirror);
	/*!
	 * \brief Part of communication at the mirror side
	 * \details Used by \ref MPIB_measure_p2p
	 */
	void (*execute_mirror)(void* this, MPI_Comm comm, int M, int measure);
	/*!
	 * \brief Finalization of data required for the communication operation
	 * \details Used by \ref MPIB_measure_p2p
	 */
	void (*finalize)(void* this, MPI_Comm comm);
	/*!
	 * \brief Frees container
	 * \details Usage:
	 * \code
MPIB_p2p_container* container = ...;
...
container->free(container);
	 * \endcode
	 */
	void (*free)(void* this);
}
MPIB_p2p_container;

/*! \brief Allocates container for send/recv. */
MPIB_p2p_container* MPIB_Send_Recv_container_alloc();

/*!
 * \brief Container for a collective communication operation to be measured by:
 * \ref MPIB_measure_max, \ref MPIB_measure_root, \ref MPIB_measure_global
 * \param this pointer to itself
 * \param comm MPI communicator over which the communication operation will be performed
 * \param root root process
 * \param M message size
 */
typedef struct MPIB_coll_container
{
	/*!
	 * \brief Initializion of data required for the communication operation
	 * \details Used by \ref MPIB_measure_coll
	 */
	void (*initialize)(void* this, MPI_Comm comm, int root, int M);
	/*!
	 * \brief Communication operation
	 * \details Used by \ref MPIB_measure_coll
	 */
	void (*execute)(void* this, MPI_Comm comm, int root, int M);
	/*!
	 * \brief Finalization of data required for the communication operation
	 * \details Used by \ref MPIB_measure_coll
	 */
	void (*finalize)(void* this, MPI_Comm comm, int root);
	/*!
	 * \brief Frees container
	 * \details Usage:
	 * \code
MPIB_p2p_container* container = ...;
...
container->free(container);
	 * \endcode
	 */
	void (*free)(void* this);
}
MPIB_coll_container;

/*!
 * \brief Allocates container for a collective operation.
 * \param operation collective operation:
 * - MPI_Scatter, MPIB_Scatter_linear, MPIB_Scatter_binomial
 * - MPI_Gather, MPIB_Gather_linear, MPIB_Gather_binomial
 * - MPI_Scatterv
 * - MPI_Gatherv
 * \return container (NULL if none of above)
 */
MPIB_coll_container* MPIB_coll_container_alloc(const char* operation);

/*!
 * \brief Allocates container for scatter.
 * \param scatter scatter implementation
 * \return container
 */
MPIB_coll_container* MPIB_Scatter_container_alloc(MPIB_Scatter scatter);

/*!
 * \brief Allocates container for gather.
 * \param gather gather implementaiton
 * \return container
 */
MPIB_coll_container* MPIB_Gather_container_alloc(MPIB_Gather gather);

/*! \brief Allocates container for scatterv. */
MPIB_coll_container* MPIB_Scatterv_container_alloc(int stride);

/*! \brief Allocates container for gatherv. */
MPIB_coll_container* MPIB_Gatherv_container_alloc(int stride);

/*!
 * \}
 */

#endif /*MPIB_CONTAINER_H_*/
