#include "mpib.h"
#include <stdio.h>
#include <malloc.h>

void MPIB_max_wtick(MPI_Comm comm, double* wtick)
{
	double tmp = MPI_Wtick();
	MPI_Allreduce(&tmp, wtick, 1, MPI_DOUBLE, MPI_MAX, comm); 
}

void MPIB_measure_p2p(MPIB_p2p_container* container, MPI_Comm comm, int M, int parallel,
	MPIB_precision precision, MPIB_result* results)
{
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	if (size < 2)
	{
		if (rank == 0)
			fprintf(stderr, "Cannot perform p2p benchmarks for %d processes (must be >= 2)\n", size);
		return;
	}

	int sendcount =  size - rank - 1;
	MPIB_result* res = (MPIB_result*)malloc(sizeof(MPIB_result) * sendcount);

	double* T = (double*)malloc(sizeof(double) * precision.max_reps);
	container->initialize(container, comm, M);
	MPIB_pairs* pairs = MPIB_build_pairs(size);
	MPIB_pairs* iter_pairs = pairs;
	while (iter_pairs)
	{
		MPI_Barrier(comm);
		MPIB_pair* iter_pair = iter_pairs->list;
		while (iter_pair)
		{
			if (!parallel)
				MPI_Barrier(comm);
			int i = iter_pair->values[0];
			int j = iter_pair->values[1];
			if (rank == i || rank == j)
			{
				int stop = 0;
				double sum = 0;
				int reps = 0;
				double ci = 0;
				while (!stop && reps < precision.max_reps)
				{
					if (rank == i)
					{
						MPI_Recv(NULL, 0, MPI_CHAR, j, 0, comm, MPI_STATUS_IGNORE);
						T[reps] = MPI_Wtime();
						container->execute_measure(container, comm, M, j);
						sum += T[reps] = MPI_Wtime() - T[reps];
						reps++;
						if (reps >= precision.min_reps && reps > 2)
						{
							stop = (ci = MPIB_ci(precision.cl, reps, T)) * reps / sum < precision.eps;
							MPI_Send(&stop, 1, MPI_INT, j, 0, comm);
						} 
					}
					if (rank == j)
					{
						MPI_Send(NULL, 0, MPI_CHAR, i, 0, comm);
						container->execute_mirror(container, comm, M, i);
						reps++;
						if (reps >= precision.min_reps && reps > 2)
							MPI_Recv(&stop, 1, MPI_INT, i, 0, comm, MPI_STATUS_IGNORE);
					}
				}
				if (rank == i)
				{
					int index = j - i - 1; 
					res[index].T = sum / reps;
					res[index].wtick = MPI_Wtick();
					res[index].reps = reps;
					res[index].ci = ci;
				}
			}
			iter_pair = iter_pair->next;
		}
		iter_pairs = iter_pairs->next;
	}
	MPIB_free_pairs(pairs);
	container->finalize(container, comm);
	free(T);

	int* recvcounts = (int*)malloc(sizeof(int) * size);
	int* displs = (int*)malloc(sizeof(int) * size);
	int i;
	for (i = 0; i < size; i++)
	{
		recvcounts[i] = sizeof(MPIB_result) * (size - i - 1);
		displs[i] = i == 0 ? 0 : displs[i - 1] + recvcounts[i - 1]; 
	}
	MPI_Allgatherv(res, sizeof(MPIB_result) * sendcount, MPI_CHAR, results, recvcounts, displs, MPI_CHAR, comm);
	free(recvcounts);
	free(displs);

	free(res);
}
