#ifndef MPIB_MEASUREMENT_H_
#define MPIB_MEASUREMENT_H_

/*!
 * \defgroup measurement Measurement
 * \brief Base structures and functions for measurements.
 * \{
 */ 

/*!
 * \brief Returns a resolution of MPI_Wtime, maximum in the communicator.
 * \details Result can be used to check the execution time measured at several processors
 * (\ref MPIB_measure_max, \ref MPIB_measure_global): \f$ T_{coll} < wtick_{max} \f$.
 * \param comm MPI communicator
 * \param wtick a maximum resolution
 */
void MPIB_max_wtick(MPI_Comm comm, double* wtick);

/*!
 * \brief Measurement presicion
 * \details To provide reliable results, the communication experiments in each benchmark are repeated
 * either fixed or variable number of times. This allows the user to control the accuracy of 
 * the obtained estimation of the execution time.
 * - Assigning to \ref min_reps and \ref max_reps the same values results in the fixed number of repetitions of
 * the communication operation, with the \ref cl argument being ignored.
 * - If \ref min_reps < \ref max_reps, the experiments are repeated until a confidence interval, ci,
 * found with the confidence level, \ref cl = \f$ Pr(|\bar T - \mu| < ci) \f$, satisfies
 * \f$ \frac{ci}{|\bar T|} < \epsilon \f$ = \ref eps, or the number of repetitions reaches its maximum,
 * \ref max_reps.
 * \note As communication operations in a series are isolated from each other, we suppose that
 * the execution times are an independent sample from a normally distributed population.
 */
typedef struct MPIB_precision
{
	/*! \brief minimum number of repetitions */
	int min_reps;
	/*! \brief maximum number of repetitions */
	int max_reps;
	/*!
	 * \brief confidence level \f$ \in [0, 1] \f$
	 * \details \f$ cl = Pr(|\bar T - \mu| < ci) = Pr(\frac{|\bar T - \mu|}{|\bar T|} < \epsilon) \f$
	 */
	double cl;
	/*!
	 * \brief error \f$ \in [0, 1] \f$
	 * \details \f$ \frac{|\bar T - \mu|}{|\bar T|} < \frac{ci}{|\bar T|} < \epsilon = eps \f$
	 */
	double eps;
}
MPIB_precision;

/*! \brief Result of the measurement defined by \ref MPIB_precision */
typedef struct MPIB_result
{
	/*! \brief execution time */
	double T;
	/*! \brief resolution of MPI_Wtime */
	double wtick;
	/*! \brief number of repetitions the benchmark has actually taken */
	int reps;
 	/*! \brief confidence interval, \f$ |\bar T - \mu| < ci \f$. */
	double ci;
}
MPIB_result;

/*!
 * \}
 * \defgroup measurement_p2p Measurement of point-to-point communications
 * \brief Measures the execution time of point-to-point communications.
 * \{
 */

/*!
 * \brief Measures the point-to-point execution time.
 * \details Measures the execution time of each \f$ i \xlongleftrightarrow[M]{M} j \f$ roundtrip
 * in the communicator, \f$ i < j \f$. Performes series of roundtrips to obtain reliable results.
 * \param container communication operaion container
 * \param comm communicator, number of nodes should be \f$ \ge 2 \f$
 * \param M message size
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param precision measurement precision
 * \param results array of \f$ C_n^2 \f$ measurement results
 */
void MPIB_measure_p2p(MPIB_p2p_container* container, MPI_Comm comm, int M, int parallel,
	MPIB_precision precision, MPIB_result* results);

/*!
 * \}
 * \defgroup measurement_coll_uni Measurement of collective communications (universal methods)
 * \brief Measures the execution time of collective communications by universal methods.
 * \{
 */

/*!
 * \brief Measures the execution time of collective operation.
 * \param container communication operaion container
 * \param comm communicator
 * \param root root process
 * \param M message size
 * \param precision measurement precision
 * \param result measurement result
 */
typedef void (*MPIB_measure_coll)(MPIB_coll_container* container, MPI_Comm comm, int root, int M,
	MPIB_precision precision, MPIB_result* result);

/*!
 * \brief Measures the execution time of collective operation at all processes and finds a maximum.
 * \details In the loop over repetitions:
 * - Synchronizes the processes by double barrier.
 * - Measures the execution time of collective operation at all processes.
 * - Finds maximum by allreduce.
 * - Performs statistical analysis.
 */
void MPIB_measure_max(MPIB_coll_container* container, MPI_Comm comm, int root, int M,
	MPIB_precision precision, MPIB_result* result);

/*!
 * \brief Measures the average execution time of barrier at all processes in the communicator
 * and sets up an internal global variable, which is used by \ref MPIB_measure_root.
 * \details Called by \ref MPIB_measure_root. Can be called directly before measurements.
 * \param comm communicator
 * \param reps number of repetitions (not precision - we suppose no pipeline effect with barrier)
 */
void MPIB_root_timer_init(MPI_Comm comm, int reps);

/*!
 * \brief Measures the execution time of collective operation at the root process.
 * \details Reuses already obtained barrier time if the previous initialization was performed
 * over the same communicator. Otherwise, initializes the root timer by calling \ref MPIB_root_timer_init.\n
 * In the loop over repetitions:
 * - Synchronizes the processes by double barrier.
 * - Measures the execution time of collective operation and barrier confirmation at the root process.
 * - Subtracts the execution time of barrier.
 * - Performs statistical analysis at root.
 * 
 * Broadcasts the result.
 */
void MPIB_measure_root(MPIB_coll_container* container, MPI_Comm comm, int root, int M,
	MPIB_precision precision, MPIB_result* result);

/*!
 * \brief Measures the offsets between local clocks of of all processes in the communicator
 * and sets up an internal global variable, which is used by \ref MPIB_measure_global.
 * \details If MPI_WTIME_IS_GLOBAL (MPI global timer) is defined, the offsets are set to zero.
 * Otherwise, the offsets are measured.
 * Called by \ref MPIB_measure_global. Can be called directly before measurements.
 * \attention As the global variable is an array, the memory should be freed by
 * \code MPIB_global_timer_init(MPI_COMM_NULL, 0) \endcode
 * \param comm communicator
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param reps number of repetitions (not precision because series of ping-pongs are required - we cannot
 * interrupt them by sending/receiving the result of the statistical estimation)
 */
void MPIB_global_timer_init(MPI_Comm comm, int parallel, int reps);

/*!
 * \brief Measures the execution time of collective operation between processes using global time.
 * \details Reuses already obtained offsets between local clocks if the previous initialization was performed
 * on the same communicator. Otherwise, initializes the global timer by calling \ref MPIB_global_timer_init.\n
 * In the loop over repetitions:
 * - Synchronizes the processes by double barrier.
 * - Measures the moment of start at the root and the moment of finish at the rest of processes.
 * - Having substructed the offset, finds maximum by reducing to the root. 
 * - Performs statistical analysis at root.
 * 
 * Broadcasts the result.
 */
void MPIB_measure_global(MPIB_coll_container* container, MPI_Comm comm, int root, int M,
	MPIB_precision precision, MPIB_result* result);

/*!
 * \}
 * \defgroup measurement_coll_spec Measurement of collective communications (operation-specific methods)
 * \brief Measures the execution time of collective communications by operation-specific methods.
 * \{
 */

/*!
 * \brief Performs the p2p benchmark with empty message with given precision
 * and sets up an internal global variable, which is used by \ref MPIB_measure_bcast.
 * \details Called by \ref MPIB_measure_bcast. Can be called directly before measurements.
 * \attention As the global variable is an array, the memory should be freed by
 * \code MPIB_bcast_timer_init(MPI_COMM_NULL, (MPIB_precision){0, 0, 0}) \endcode
 * \param comm communicator
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param precision measurement precision
 */
void MPIB_bcast_timer_init(MPI_Comm comm, int parallel, MPIB_precision precision);

/*!
 * \brief Measures the execution time of bcast between root and the rest of processes.
 * \details Reuses already obtained empty-roundtrip times if the previous initialization was performed
 * on the same communicator. Otherwise, initializes the bcast timer by calling \ref MPIB_bcast_timer_init.\n
 * In the loop over processes, except root:\n
 * In the loop over repetitions:
 * - bcast at all processes except root and current process in the loop
 * - bcast and empty recv at root and measures the execution time
 * - bcast and empty send at current process in the loop
 * 
 * Having substructed the half of empty-roundtrip times, finds maximum.\n 
 * Broadcasts the result.
 * \note No double barriers as no need in synchronization.
 * \note No precision as we cannot interrupt the process of measurement by broadcasting the error value
 * after each repetition.
 * \param bcast bcast implementation
 * \param comm communicator
 * \param root root process
 * \param M message size
 * \param max_reps maximum number of repetitions
 * \param result measurement result
 */
void MPIB_measure_bcast(MPIB_Bcast bcast, MPI_Comm comm, int root, int M,
	int max_reps, MPIB_result* result);

/*!
 * \}
 */

#endif /*MPIB_MEASUREMENT_H_*/
