#ifndef MPIB_UTILITIES_H_
#define MPIB_UTILITIES_H_

/*!
 * \defgroup utilities Utilities
 * \{
 */

/*!
 * \brief Returns a confidence interval that contains the average execution time with a certain probability.
 * \details \f$ Pr(|\bar T - \mu| < ci) = cl \f$
 * \param cl confidence level
 * \param reps number of measurements (should be > 1)
 * \param T array of reps measurement results 
 * \return confidence interval
 */
double MPIB_ci(double cl, int reps, double* T);

/*! \brief \f$ C_n^2 \f$ */
#define MPIB_C2(n) (n) * ((n) - 1) / 2

/*!
 * \brief For a symmetric square matrix stored in the array of \f$ C_n^2 \f$ elements,
 * returns the index of the \f$ (i, j) \f$ element, \f$ i \ne j < n \f$.
 * \details \f$ \frac{(n - 1) + (n - I)}{2} I + (J - I - 1) \f$, \f$ I = min(i, j) \f$, \f$ J = max(i, j) \f$
 */
#define MPIB_IJ2INDEX(n, i, j) (2 * (n) - ((i) < (j) ? (i) : (j)) - 1) * (((i) < (j) ? (i) : (j))) / 2 + (((i) < (j) ? (j) : (i))) - (((i) < (j) ? (i) : (j))) - 1

/*! \brief List of pairs */
typedef struct MPIB_pair
{
	/*! \brief values */
	int values[2];
	/*! \brief previous pair */
	struct MPIB_pair* prev;
	/*! \brief next pair */
	struct MPIB_pair* next;
}
MPIB_pair;

/*! \brief List of lists of pairs */
typedef struct MPIB_pairs
{
	/*! \brief items */
	MPIB_pair* list;
	/*! \brief previous list */
	struct MPIB_pairs* prev;
	/*! \brief next list */
	struct MPIB_pairs* next;
}
MPIB_pairs;

/*! \brief Builds all combinations of non-overlapped pairs of communicating nodes. */
MPIB_pairs* MPIB_build_pairs(int n);

/*! \brief Frees the pairs recursively. */
void MPIB_free_pairs(MPIB_pairs* pairs);

/*! \brief Prints a precision */
void MPIB_print_precision(MPIB_precision precision);

/*!
 * \brief Prints the per-p2p table of the p2p benchmark results.
 * \param precision precision
 * \param M message size
 * \param n number of nodes
 * \param results array of \f$ C_n^2 \f$ results
 */
void MPIB_print_p2p_table(int M, MPIB_precision precision, int n, MPIB_result* results);

/*!
 * \brief Prints the table header of the p2p benchmark results.
 * \param precision precision
 * \param n number of nodes
 */
void MPIB_print_p2p_th(MPIB_precision precision, int n);

/*!
 * \brief Prints the table row of the p2p benchmark results.
 * \param M message size
 * \param n number of nodes
 * \param results array of \f$ C_n^2 \f$ results
 */
void MPIB_print_p2p_tr(int M, int n, MPIB_result* results);

/*!
 * \brief Prints the table header of the collective benchmark result.
 * \details If operation == NULL, header is commented, with the time coulumn title "time".
 * Otherwise, header is uncommented (for gnuplot autoheader), with the time column title operation.
 * \param operation collective operation
 * \param n number of nodes
 * \param root root process
 * \param precision precision
 */
void MPIB_print_coll_th(const char* operation, int n, int root, MPIB_precision precision);

/*!
 * \brief Prints the table row of the collective benchmark result.
 * \param M message size
 * \param result measurement result
 */
void MPIB_print_coll_tr(int M, MPIB_result result);

/*!
 * \}
 */

#endif /*MPIB_UTILITIES_H_*/
