/*!
 * \file
 * \brief Collective benchmark
 * \details Checks and benchmarks collective operations
 * 
 * Plot can be made by \b MPIBlib/collective/collective.plot
 */

#include "cpm.h"
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

int main(int argc, char** argv)
{
	MPI_Init(&argc, &argv);

	int rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	int size;
	MPI_Comm_size(MPI_COMM_WORLD, &size);

	int exit = 0;
	char operation[256] = "";
	MPIB_coll_container* container = NULL;
	char timing[256] = "max";
	MPIB_measure_coll measure = MPIB_measure_max;
	int stride = 1024;
	int max_size = 102400;
	int parallel = 1;
	MPIB_precision precision = {5, 100, 0.95, 0.025};
	char* filename;// at 0
	int build = 1;

	if (rank == 0)
	{
		int c;
		while ((c = getopt(argc, argv, "hO:t:s:m:pr:R:c:e:i:")) != -1)
		{
			switch (c)
			{
				case 'h':
					fprintf
					(
						stderr,
"usage: mpirun [mpirun options] %s [options]\n\
-h			help\n\
-O S			collective operation (required):\n\
			MPI_Scatter, MPIB_Scatter_linear, MPIB_Scatter_binomial\n\
			MPI_Gather, MPIB_Gather_linear, MPIB_Gather_binomial\n\
			MPI_Scatterv\n\
			MPI_Gatherv\n\
			CPM_Scatter_opt\n\
			CPM_Gather_opt\n\
			CPM_Scatter_binomial_opt, CPM_Scatter_binomial_cor\n\
			CPM_Gather_binomial_opt, CPM_Gather_binomial_cor\n\
-t S			timing: max, root, global (default: %s)\n\
-s I			message size stride (default: %d)\n\
-m I			maximum message size (default: %d)\n\
-p			parallel p2p benchmarking (default: %d)\n\
-r I			minimum number of repetitions (default: %d)\n\
-R I			maximum number of repetitions (default: %d)\n\
-c D			confidence level: 0 < D < 1 (default: %le)\n\
-e D			error: 0 < D < 1 (default: %le)\n\
-i S			input model file, stdin - standard input (default: none - model will be built)\n",
						argv[0],
						timing,
						stride,
						max_size,
						parallel,
						precision.min_reps,
						precision.max_reps,
						precision.cl,
						precision.eps
					);
					exit = 1;
					break;
				case 'O':
					strcpy(operation, optarg);
					break;
				case 't':
					strcpy(timing, optarg);
					break;
				case 's':
					stride = atoi(optarg);
					break;
				case 'm':
					max_size = atoi(optarg);
					break;
				case 'p':
					parallel = 1;
					break;
				case 'r':
					precision.min_reps = atoi(optarg);
					if (precision.max_reps < precision.min_reps)
						precision.max_reps = precision.min_reps;
					break;
				case 'R':
					precision.max_reps = atoi(optarg);
					if (precision.min_reps > precision.max_reps)
						precision.min_reps = precision.max_reps;
					break;
				case 'c':
					precision.cl = atof(optarg);
					break;
				case 'e':
					precision.eps = atof(optarg);
					break;
				case 'i':
					filename = optarg;
					build = 0;
					break;
			}
		}
	}
	MPI_Bcast(&exit, 1, MPI_INT, 0, MPI_COMM_WORLD);
	if (exit)
		goto exit;
	MPI_Bcast(operation, 256, MPI_CHAR, 0, MPI_COMM_WORLD);
	container = MPIB_coll_container_alloc(operation);
	if (!container)
	{
		if (rank == 0)
			fprintf(stderr, "Collective operation %s undefined\n", operation);
		goto exit;
	}
	MPI_Bcast(timing, 256, MPI_CHAR, 0, MPI_COMM_WORLD);
	if (!strcmp(timing, "root"))
		measure = MPIB_measure_root;
	if (!strcmp(timing, "global"))
		measure = MPIB_measure_global;
	MPI_Bcast(&stride, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&max_size, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&precision, sizeof(MPIB_precision), MPI_CHAR, 0, MPI_COMM_WORLD);
	MPI_Bcast(&build, 1, MPI_INT, 0, MPI_COMM_WORLD);

	if (!strcmp(operation, "CPM_Scatter_opt") || !strcmp(operation, "CPM_Gather_opt"))
	{
		CPM_model* model = NULL;// at 0
		if (build)
			CPM_build(MPI_COMM_WORLD, stride, max_size, parallel, precision, &model);
		else
		{
			if (rank == 0)
			{
				model = CPM_model_load(filename);
				exit = model == NULL;
			}
		}
		MPI_Bcast(&exit, 1, MPI_INT, 0, MPI_COMM_WORLD);
		if (exit)
			goto exit;
		CPM_model_initialize(MPI_COMM_WORLD, model);
		if (rank == 0)
			CPM_model_free(model);
		if (!strcmp(operation, "CPM_Scatter_opt")) 
			container = MPIB_Scatter_container_alloc(CPM_Scatter_opt);
		if (!strcmp(operation, "CPM_Gather_opt"))
			container = MPIB_Gather_container_alloc(CPM_Gather_opt);
	}
	if (!strcmp(operation, "CPM_Scatter_binomial_opt") || !strcmp(operation, "CPM_Scatter_binomial_cor") ||
		!strcmp(operation, "CPM_Gather_binomial_opt") || !strcmp(operation, "CPM_Gather_binomial_cor"))
	{
		Hockney_model* model = NULL;
		if (build)
			Hockney_build(MPI_COMM_WORLD, max_size, parallel, precision, &model); 
		else
		{
			if (rank == 0)
			{
				model = Hockney_load(filename);
				exit = model == NULL;
			}
		}
		MPI_Bcast(&exit, 1, MPI_INT, 0, MPI_COMM_WORLD);
		if (exit)
			goto exit;
		Hockney_initialize(MPI_COMM_WORLD, model);
		if (rank == 0)
			Hockney_free(model);
		if (!strcmp(operation, "CPM_Scatter_binomial_opt"))
			container = MPIB_Scatter_container_alloc(CPM_Scatter_binomial_opt);
		if (!strcmp(operation, "CPM_Scatter_binomial_cor"))
			container = MPIB_Scatter_container_alloc(CPM_Scatter_binomial_cor);
		if (!strcmp(operation, "CPM_Gather_binomial_opt"))
			container = MPIB_Gather_container_alloc(CPM_Gather_binomial_opt);
		if (!strcmp(operation, "CPM_Gather_binomial_cor"))
			container = MPIB_Gather_container_alloc(CPM_Gather_binomial_cor);
	}

	if (strstr(operation, "MPIB_Scatter_") || strstr(operation, "MPIB_Gather_") ||
		strstr(operation, "CPM_Scatter_") || strstr(operation, "CPM_Gather_"))
	{
		if (rank == 0)
			printf("#Test of %s\n", operation);
		int root;
		for (root = 0; root < size; root++)
		{
			int res;
			if (strstr(operation, "MPIB_Scatter_") ||
				strstr(operation, "CPM_Scatter_"))
			{
				MPIB_Scatter* scatter = (MPIB_Scatter*)((void*)container + sizeof(MPIB_coll_container) + sizeof(char*));
				MPIB_test_scatter(*scatter, MPI_COMM_WORLD, root, &res);
			}
			if (strstr(operation, "MPIB_Gather_") ||
				strstr(operation, "CPM_Gather_"))
			{
				MPIB_Gather* gather = (MPIB_Gather*)((void*)container + sizeof(MPIB_coll_container) + sizeof(char*));
				MPIB_test_gather(*gather, MPI_COMM_WORLD, root, &res);
			}
			if (rank == 0)
				printf("#root: %d\tresult: %d\n", root, res);
		}
	}

	if (rank == 0)
	{
		printf("#%s, %s timing\n", operation, timing);
		MPIB_print_coll_th(operation, size, 0, precision);
	}
	double time = 0;
	int M;
	for (M = 0; M < max_size; M += stride)
	{
		MPIB_result result;
		int start = MPI_Wtime();
		measure(container, MPI_COMM_WORLD, 0, M, precision, &result);
		time += MPI_Wtime() - start;
		if (rank == 0)
			MPIB_print_coll_tr(M, result);
	}
	if (rank == 0)
		printf("#total benchmarking time: %le\n", time);

	if (!strcmp(operation, "CPM_Scatter_opt") || !strcmp(operation, "CPM_Gather_opt"))
		CPM_model_finalize(MPI_COMM_WORLD);

	if (!strcmp(operation, "CPM_Scatter_binomial_opt") || !strcmp(operation, "CPM_Scatter_binomial_cor") ||
		!strcmp(operation, "CPM_Gather_binomial_opt") || !strcmp(operation, "CPM_Gather_binomial_cor"))
		Hockney_finalize(MPI_COMM_WORLD);

	container->free(container);
exit:
	MPI_Finalize();
	return 0;
}
