/*!
 * \file
 * \brief Estimates the execution time of p2p and collective operations according to the model.
 * \details The model is built or read from file.
 * 
 * \b model.plot draws the graph of the execution time (sec) against message size (kb) with error bars:
 * observation and prediction.
 * - input: model.out
 * - output: scatter.eps (MPIB_Scatter_linear), gather.eps (MPIB_Gather_linear), p2p.eps (0-1)
 */

#include "cpm.h"
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>

int main(int argc, char** argv)
{
	MPI_Init(&argc, &argv);

	int rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	int size;
	MPI_Comm_size(MPI_COMM_WORLD, &size);

	int exit = 0;
	CPM_model* model = NULL;// at 0
	int build = 1;
	char* out = "stdout";// at 0
	int stride = 1024;
	int max_size = 102400;
	int parallel = 1;
	MPIB_precision precision = {5, 100, 0.95, 0.025};

	if (rank == 0)
	{
		int c;
		while ((c = getopt(argc, argv, "hi:o:s:m:pr:R:c:e:")) != -1)
		{
			switch (c)
			{
				case 'h':
					fprintf
					(
						stderr,
"usage: mpirun [mpirun options] %s [options]\n\
-h			help\n\
-i S			input model file, stdin - standard input (default: none - model will be built)\n\
-o S			output model file, stdout - standard output (default: stdout)\n\
-s N			stride for message sizes (default: %d)\n\
-m N			maximum message size (default: %d)\n\
-p			parallel p2p benchmarking (default: %d)\n\
-r N			minimum number of repetitions (default: %d)\n\
-R N			maximum number of repetitions (default: %d)\n\
-c D			confidence level: 0 < D < 1 (default: %le)\n\
-e D			error: 0 < D < 1 (default: %le)\n",
						argv[0],
						stride,
						max_size,
						parallel,
						precision.min_reps,
						precision.max_reps,
						precision.cl,
						precision.eps
					);
					exit = 1;
					break;
				case 'i':
					model = CPM_model_load(optarg);
					exit = model == NULL;
					build = 0;
					break;
				case 'o':
					out = optarg;
					break;
				case 's':
					stride = atoi(optarg);
					break;
				case 'm':
					max_size = atoi(optarg);
					break;
				case 'p':
					parallel = 1;
					break;
				case 'r':
					precision.min_reps = atoi(optarg);
					if (precision.max_reps < precision.min_reps)
						precision.max_reps = precision.min_reps;
					break;
				case 'R':
					precision.max_reps = atoi(optarg);
					if (precision.min_reps > precision.max_reps)
						precision.min_reps = precision.max_reps;
					break;
				case 'c':
					precision.cl = atof(optarg);
					break;
				case 'e':
					precision.eps = atof(optarg);
					break;
			}
		}
	}
	MPI_Bcast(&exit, 1, MPI_INT, 0, MPI_COMM_WORLD);
	if (exit)
		goto exit;
	MPI_Bcast(&build, 1, MPI_INT, 0, MPI_COMM_WORLD);
	if (build)
	{
		MPI_Bcast(&stride, 1, MPI_INT, 0, MPI_COMM_WORLD);
		MPI_Bcast(&max_size, 1, MPI_INT, 0, MPI_COMM_WORLD);
		MPI_Bcast(&parallel, 1, MPI_INT, 0, MPI_COMM_WORLD);
		MPI_Bcast(&precision, sizeof(MPIB_precision), MPI_CHAR, 0, MPI_COMM_WORLD);

		if (rank == 0)
		{
			exit = CPM_initR();
			if (exit)
				fprintf(stderr, "R cannot be initialized\n");
		}
		MPI_Bcast(&exit, 1, MPI_INT, 0, MPI_COMM_WORLD);
		if (exit)
			goto exit;

		double time = MPI_Wtime();
		CPM_build(MPI_COMM_WORLD, stride, max_size, parallel, precision, &model);
		if (rank == 0)
		{
			printf("#CPM building time: %le\n", MPI_Wtime() - time);
			CPM_endR();
		}
	}
	if (rank == 0)
	{
		CPM_model_save(out, model);
		printf("\n\n");
	}

	char* operation;
	MPIB_coll_container* coll_container;
	const char* timing = "max";
	MPIB_measure_coll measure = MPIB_measure_max;

	operation = "MPIB_Scatter_linear";
	coll_container = MPIB_coll_container_alloc(operation);
	if (rank == 0)
	{
		printf("#%s, %s timing\n", operation, timing);
		MPIB_print_coll_th(NULL, size, 0, precision);
	}
	int M;
	for (M = 0; M < max_size; M += stride)
	{
		MPIB_result result;
		measure(coll_container, MPI_COMM_WORLD, 0, M, precision, &result);
		if (rank == 0)
			MPIB_print_coll_tr(M, result);
	}
	if (rank == 0)
		printf("\n\n");
	coll_container->free(coll_container);

	operation = "MPIB_Gather_linear";
	coll_container = MPIB_coll_container_alloc(operation);
	if (rank == 0)
	{
		printf("#%s, %s timing\n", operation, timing);
		MPIB_print_coll_th(NULL, size, 0, precision);
	}
	for (M = 0; M < max_size; M += stride)
	{
		MPIB_result result;
		measure(coll_container, MPI_COMM_WORLD, 0, M, precision, &result);
		if (rank == 0)
			MPIB_print_coll_tr(M, result);
	}
	if (rank == 0)
		printf("\n\n");
	coll_container->free(coll_container);

	MPIB_p2p_container* p2p_container = MPIB_Send_Recv_container_alloc();
	MPIB_result* results = (MPIB_result*)malloc(sizeof(MPIB_result) * MPIB_C2(size));
	if (rank == 0)
	{
		printf("#%s p2p benchmark\n", parallel ? "parallel" : "sequential");
		MPIB_print_p2p_th(precision, size);
	}
	for (M = 0; M < max_size; M += stride)
	{
		MPIB_measure_p2p(p2p_container, MPI_COMM_WORLD, M, parallel, precision, results);
		if (rank == 0)
			MPIB_print_p2p_tr(M, size, results);
	}
	if (rank == 0)
		printf("\n\n");
	free(results);
	p2p_container->free(p2p_container);

	if (rank == 0)
	{
		int M;
		printf("#scatter prediction\n");
		printf("msg\tprediction\n");
		M = 0;
		printf("%d\t%le\n", M, CPM_estimate_one2many(model, 0, M));
		M = model->S - 1;
		printf("%d\t%le\n", M, CPM_estimate_one2many(model, 0, M));
		M = model->S + 1;
		printf("%d\t%le\n", M, CPM_estimate_one2many(model, 0, M));
		M = max_size;
		printf("%d\t%le\n", M, CPM_estimate_one2many(model, 0, M));
		printf("\n\n");
		printf("#gather prediction\n");
		printf("msg\tprediction\n");
		M = 0;
		printf("%d\t%le\n", M, CPM_estimate_many2one(model, 0, M));
		M = model->M2 - 1;
		printf("%d\t%le\n", M, CPM_estimate_many2one(model, 0, M));
		M = model->M2 + 1;
		printf("%d\t%le\n", M, CPM_estimate_many2one(model, 0, M));
		M = max_size;
		printf("%d\t%le\n", M, CPM_estimate_many2one(model, 0, M));
		printf("\n\n");
		printf("#p2p prediction\n");
		printf("#msg\t");
		int i;
		for(i = 0; i < size - 1; i++)
		{
			int j;
			for(j = i + 1; j < size; j++)
				printf("%d-%d\t\t", i, j);
		}
		printf("\n");
		printf("0");
		for(i = 0; i < size - 1; i++)
		{
			int j;
			for(j = i + 1; j < size; j++)
				printf("\t%le", CPM_estimate_p2p(model, i, j, 0));
		}
		printf("\n");
		printf("%d", max_size);
		for(i = 0; i < size - 1; i++)
		{
			int j;
			for(j = i + 1; j < size; j++)
				printf("\t%le", CPM_estimate_p2p(model, i, j, max_size));
		}
		printf("\n");

		CPM_model_free(model);
	}

exit:
	MPI_Finalize();
	return 0;
}
