#include "cpm.h"
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <float.h>
#include <gsl/gsl_heapsort.h>

const char* message = "model instance must be initialized\n";

CPM_model* CPM_model_instance = NULL;

void CPM_model_initialize(MPI_Comm comm, CPM_model* model)
{
	int rank;
	MPI_Comm_rank(comm, &rank);
	int n = 0;
	if (rank == 0)
		n = model->n;
	MPI_Bcast(&n, 1, MPI_INT, 0, comm); 
	CPM_model_instance = CPM_model_alloc(n);
	if (rank == 0)
		memcpy(CPM_model_instance->C, model->C, sizeof(double) * n);
	MPI_Bcast(CPM_model_instance->C, n, MPI_DOUBLE, 0, comm);
	if (rank == 0)
		memcpy(CPM_model_instance->t, model->t, sizeof(double) * n);
	MPI_Bcast(CPM_model_instance->t, n, MPI_DOUBLE, 0, comm);
	int Cn2 = n * (n - 1) / 2; 
	if (rank == 0)
		memcpy(CPM_model_instance->b, model->b, sizeof(double) * Cn2);
	MPI_Bcast(CPM_model_instance->b, Cn2, MPI_DOUBLE, 0, comm);
	if (rank == 0)
		CPM_model_instance->S = model->S; 
	MPI_Bcast(&CPM_model_instance->S, 1, MPI_DOUBLE, 0, comm);
	if (rank == 0)
		CPM_model_instance->M1 = model->M1; 
	MPI_Bcast(&CPM_model_instance->M1, 1, MPI_DOUBLE, 0, comm);
	if (rank == 0)
		CPM_model_instance->M2 = model->M2; 
	MPI_Bcast(&CPM_model_instance->M2, 1, MPI_DOUBLE, 0, comm);
}

void CPM_model_finalize(MPI_Comm comm)
{
	CPM_model_free(CPM_model_instance);
	CPM_model_instance = NULL;
}

int CPM_Scatter_opt(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype, 
	int root, MPI_Comm comm)
{
	if (CPM_model_instance != NULL)
	{
		int type_size;
		MPI_Type_size(sendtype, &type_size);
		int S = CPM_model_instance->S;
		int sz = sendcount * type_size; 
		if (sz > S)
		{
			int rank;
			MPI_Comm_rank(comm, &rank);
			int size;
			MPI_Comm_size(comm, &size);
	
			int m = 1;
			while (sz / m >= S || sz / (m - 1) < S) 
				m++;
			int count = sz / m;
			char* tmp = rank == root ? (char*)malloc(size * count) : NULL;
	
			int pos = 0;
			while (pos < sz)
			{
				int cnt = sz - pos < count ? sz - pos : count;
				if (rank == root)
				{
					int i;
					for (i = 0; i < size; i++)
						memcpy(tmp + cnt * i, sendbuf + sz * i + pos, cnt);
				}
				int err = MPI_Scatter(tmp, cnt, MPI_CHAR, recvbuf + pos, cnt, MPI_CHAR, root, comm);
				if (err != MPI_SUCCESS)
					return err;
				pos += cnt;
			}
	
			if (tmp != NULL)
				free(tmp);
	
			return MPI_SUCCESS;
		}
		else
			return MPI_Scatter(sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
	}
	else
	{
		fprintf(stderr, message);
		return MPI_Scatter(sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
	}
}

int CPM_Gather_opt(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype, 
	int root, MPI_Comm comm)
{
	if (CPM_model_instance != NULL)
	{
		int type_size;
		MPI_Type_size(sendtype, &type_size);
		int M1 = CPM_model_instance->M1;
		int M2 = CPM_model_instance->M2;
		int sz = sendcount * type_size; 
		if (sz > M1 && sz < M2)
		{
			int rank;
			MPI_Comm_rank(comm, &rank);
			int size;
			MPI_Comm_size(comm, &size);
	
			int m = 1;
			while (sz / m >= M1 || sz / (m - 1) < M1) 
				m++;
			int count = sz / m;
			char* tmp = rank == root ? (char*)malloc(size * count) : NULL;
	
			int pos = 0;
			while (pos < sz)
			{
				int cnt = sz - pos < count ? sz - pos : count;
				MPI_Barrier(comm);
				int err = MPI_Gather(sendbuf + pos, cnt, MPI_CHAR, tmp, cnt, MPI_CHAR, root, comm);
				if (err != MPI_SUCCESS)
					return err;
				if (rank == root)
				{
					int i;
					for (i = 0; i < size; i++)
						memcpy(recvbuf + sz * i + pos, tmp + cnt * i, cnt);
				}
				pos += cnt;
			}
	
			if (tmp != NULL)
				free(tmp);
	
			return MPI_SUCCESS;
		}
		else
			return MPI_Gather(sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
	}
	else
	{
		fprintf(stderr, message);
		return MPI_Gather(sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
	}
}

Hockney_model* Hockney_model_instance = NULL;

void Hockney_initialize(MPI_Comm comm, Hockney_model* model)
{
	int rank;
	MPI_Comm_rank(comm, &rank);
	int n = rank == 0 ? model->n : 0;
	MPI_Bcast(&n, 1, MPI_INT, 0, comm); 
	Hockney_model_instance = Hockney_alloc(n);
	int Cn2 = n * (n - 1) / 2;
	if (rank == 0)
		memcpy(Hockney_model_instance->a, model->a, sizeof(double) * Cn2);
	MPI_Bcast(Hockney_model_instance->a, Cn2, MPI_DOUBLE, 0, comm);
	if (rank == 0)
		memcpy(Hockney_model_instance->b, model->b, sizeof(double) * Cn2);
	MPI_Bcast(Hockney_model_instance->b, Cn2, MPI_DOUBLE, 0, comm);
}

void Hockney_finalize(MPI_Comm comm)
{
	Hockney_free(Hockney_model_instance);
	Hockney_model_instance = NULL;
}

typedef struct CPM_tree
{
	int rank;
	int block;
	struct CPM_tree* parent;
	int count;
	struct CPM_tree** children;
}
CPM_tree;

void CPM_tree_free(CPM_tree* tree)
{
	int i;
	for (i = 0; i < tree->count; i++)
		CPM_tree_free(tree->children[i]);
	free(tree->children);
	free(tree);
}

CPM_tree* CPM_tree_build(CPM_tree* parent, int size)
{
	if (size == 0)
		return NULL;
	CPM_tree* tree = malloc(sizeof(CPM_tree));
	tree->rank = -1;
	tree->block = size;
	tree->parent = parent;
	tree->count = 0;
	tree->children = NULL;
	int tmp, bin;
	for (tmp = size - 1, bin = 1; tmp > 0; tmp /= 2, bin *= 2)
	{
		if (tmp % 2)
		{
			tree->children = realloc(tree->children, sizeof(CPM_tree*) * (tree->count + 1));
			tree->children[tree->count] = CPM_tree_build(tree, bin);
			tree->count++;
		}
	}
	return tree;
}

void CPM_tree_get(CPM_tree* tree, int block, int* count, CPM_tree*** items)
{
	if (tree->block == block)
	{
		(*items) = realloc(*items, sizeof(CPM_tree**) * (*count + 1));
		(*items)[*count] = tree;
		(*count)++;
	}
	else
	{
		int i;
		for (i = tree->count - 1; i >= 0; i--)
			if (tree->children[i]->block >= block)
				CPM_tree_get(tree->children[i], block, count, items);
	}
}

CPM_tree* CPM_tree_find(CPM_tree* tree, int rank)
{
	if (tree->rank == rank)
		return tree;
	int i;
	for (i = 0; i < tree->count; i++)
	{
		CPM_tree* item = CPM_tree_find(tree->children[i], rank);
		if (item != NULL)
			return item;
	}
	return NULL;
}

void CPM_tree_fill_opt(CPM_tree* tree, int root)
{
	tree->rank = root;
	int bin = 1;
	while (bin < tree->block)
		bin *= 2;
	bin /= 2;
	while (bin > 0)
	{
		int count = 0;
		CPM_tree** items = NULL;
		CPM_tree_get(tree, bin, &count, &items);
		int i;
		for (i = 0; i < count; i++)
		{
			CPM_tree* item = items[i];
			double min = DBL_MAX;
			int min_index = -1;
			int j;
			for (j = 0; j < tree->block; j++)
			{
				if (!CPM_tree_find(tree, j))
				{
					double tmp = Hockney_estimate(Hockney_model_instance, item->parent->rank, j, bin);
//					double tmp = MPIB_procs_instance->a[item->parent->rank] + MPIB_procs_instance->a[j]
//						+ (MPIB_procs_instance->b[item->parent->rank] + MPIB_procs_instance->b[j]) * bin;
					if (tmp < min)
					{
						min = tmp;
						min_index = j;
					}
				} 
			}
			item->rank = min_index;
		}
		free(items);
		bin /= 2;
	}
}

void binomial_tree_opt(int size, int root, int rank, int* block, int* parent, int* count, int** children, int** blocks)
{
	CPM_tree* tree = CPM_tree_build(NULL, size);
	CPM_tree_fill_opt(tree, root);
	CPM_tree* item = CPM_tree_find(tree, rank);
	*block = item->block;
	*parent = item->parent == NULL ? -1 : item->parent->rank;
	*count = item->count;
	*children = malloc(sizeof(int) * item->count);
	*blocks = malloc(sizeof(int) * item->count);
	int i;
	for (i = 0; i < item->count; i++)
	{
		(*children)[i] = item->children[i]->rank;
		(*blocks)[i] = item->children[i]->block;
	} 
	CPM_tree_free(tree);
/*
	printf("#rank: %d\tblock: %d\tparent: %d\tchildren:", rank, *block, *parent);
	for (i = 0; i < *count; i++)
		printf(" %d", (*children)[i]);
	printf("\n");
*/
}

void CPM_tree_fill_cor(CPM_tree* tree, int root)
{
	tree->rank = root;
	int bin = 1;
	while (bin < tree->block)
		bin *= 2;
	bin /= 2;
	while (bin > 0)
	{
		int count = 0;
		CPM_tree** items = NULL;
		CPM_tree_get(tree, bin, &count, &items);
		int i;
		for (i = 0; i < count; i++)
		{
			CPM_tree* item = items[i];
			double max = 0;
			int max_index = -1;
			int j;
			for (j = 0; j < tree->block; j++)
			{
				if (!CPM_tree_find(tree, j))
				{
					double tmp = Hockney_estimate(Hockney_model_instance, item->parent->rank, j, bin);
//					double tmp = MPIB_procs_instance->a[item->parent->rank] + MPIB_procs_instance->a[j]
//						+ (MPIB_procs_instance->b[item->parent->rank] + MPIB_procs_instance->b[j]) * bin;
					if (tmp > max)
					{
						max = tmp;
						max_index = j;
					}
				} 
			}
			item->rank = max_index;
		}
		free(items);
		bin /= 2;
	}
}

void binomial_tree_cor(int size, int root, int rank, int* block, int* parent, int* count, int** children, int** blocks)
{
	CPM_tree* tree = CPM_tree_build(NULL, size);
	CPM_tree_fill_cor(tree, root);
	CPM_tree* item = CPM_tree_find(tree, rank);
	*block = item->block;
	*parent = item->parent == NULL ? -1 : item->parent->rank;
	*count = item->count;
	*children = malloc(sizeof(int) * item->count);
	*blocks = malloc(sizeof(int) * item->count);
	int i;
	for (i = 0; i < item->count; i++)
	{
		(*children)[i] = item->children[i]->rank;
		(*blocks)[i] = item->children[i]->block;
	} 
	CPM_tree_free(tree);
} 

int CPM_Scatter_binomial_opt(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype, 
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance != NULL)
//	if (MPIB_procs_instance != NULL)
		return MPIB_Scatter_tree(binomial_tree_opt, sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
	else
		return MPIB_Scatter_binomial(sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int CPM_Scatter_binomial_cor(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype, 
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance != NULL)
//	if (MPIB_procs_instance != NULL)
		return MPIB_Scatter_tree(binomial_tree_cor, sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
	else
		return MPIB_Scatter_binomial(sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int CPM_Gather_binomial_opt(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance != NULL)
//	if (MPIB_procs_instance != NULL)
		return MPIB_Gather_tree(binomial_tree_opt, sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
	else
		return MPIB_Gather_binomial(sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int CPM_Gather_binomial_cor(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance != NULL)
//	if (MPIB_procs_instance != NULL)
		return MPIB_Gather_tree(binomial_tree_cor, sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
	else
		return MPIB_Gather_binomial(sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

MPI_Comm CPM_comm_instance = MPI_COMM_NULL;

int ascending(const void* a, const void* b)
{
	double a_ = *(double*)a;
	double b_ = *(double*)b;
	if (a_ > b_)
		return 1;
	else if (a_ < b_)
		return -1;
	else
		return 0;
}

int descending(const void* a, const void* b)
{
	return -ascending(a, b);
}

void CPM_p2p_initialize(MPI_Comm comm, int root, int opt)
{
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);

	MPIB_p2p_container* container = MPIB_Send_Recv_container_alloc();
	int M = 10240;
	MPIB_precision precision = { 10, 10, 0.05 };
	MPIB_result* res = rank == 0 ? (MPIB_result*)malloc(sizeof(MPIB_result) * MPIB_C2(size)) : NULL; 
	MPIB_measure_p2p(container, comm, M, 0, precision, res);
	container->free(container);
	
	size_t* indices = malloc(sizeof(size_t) * size);
	indices[0] = root;
	if (rank == 0)
	{
		double* values = (double*)calloc((size - 1), sizeof(double));
		int i, j, index;
		for (i = 0, index = 0; i < size; i++)
		{
			for (j = 0; j < size; j++, index++)
			{
				if (i != root && j != i)
					values[i < root ? i : i - 1] += res[index].T;
			}
		}
		gsl_heapsort_index(&indices[1], values, size - 1, sizeof(double), opt ? ascending : descending);
		free(values);
		for (i = 1; i < size; i++)
		{
			if (indices[i] >= root)
				indices[i]++;
		}
	}
	MPI_Bcast(indices, size, MPI_INT, 0, comm);
	free(res);

	MPI_Group group;
	MPI_Comm_group(comm, &group);
	MPI_Group new_group;
	MPI_Group_incl(group, size, indices, &new_group);
	MPI_Group_free(&group);
	free(indices);
	MPI_Comm_create(comm, new_group, &CPM_comm_instance);
	MPI_Group_free(&new_group);
}

void CPM_p2p_finalize(MPI_Comm comm)
{
	MPI_Comm_free(&CPM_comm_instance);
}

int CPM_Scatter_binomial(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	void* recvbuf, int recvcount, MPI_Datatype recvtype, 
	int root, MPI_Comm comm)
{
	if (CPM_comm_instance != MPI_COMM_NULL)
		return MPIB_Scatter_binomial(sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, 0, CPM_comm_instance);
	else
		return MPIB_Scatter_binomial(sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}
