#include "cpm.h"
#include <stdio.h>
#include <malloc.h>
#include <memory.h>

Hockney_model* Hockney_alloc(int n)
{
	Hockney_model* model = (Hockney_model*)malloc(sizeof(Hockney_model));
	model->n = n;
	model->a = (double*)malloc(sizeof(double) * MPIB_C2(n));
	model->b = (double*)malloc(sizeof(double) * MPIB_C2(n));
	return model;
}

void Hockney_build(MPI_Comm comm, int M, int parallel, MPIB_precision precision, Hockney_model** model_)
{
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	if (size < 2)
	{
		if (rank == 0)
			fprintf(stderr, "Cannot compute Hockney parameters for %d processes (must be >= 2)\n", size);
		return;
	}
	Hockney_model* model = rank == 0 ? Hockney_alloc(size) : NULL;
	MPIB_result* res_0 = rank == 0 ? malloc(sizeof(MPIB_result) * MPIB_C2(size)) : NULL;
	MPIB_result* res_M = rank == 0 ? malloc(sizeof(MPIB_result) * MPIB_C2(size)) : NULL;
	MPIB_p2p_container* container = MPIB_Send_Recv_container_alloc();
	MPIB_measure_p2p(container, comm, 0, parallel, precision, res_0);
	MPIB_measure_p2p(container, comm, M, parallel, precision, res_M);
	if (rank == 0)
	{
		int i, n;
		for (i = 0, n = size * (size - 1) / 2; i < n; i++)
		{
			model->a[i] = res_0[i].T / 2;
			model->b[i] = (res_M[i].T - res_0[i].T) / (2 * M);
		}
		*model_ = model;
	}
	free(res_0);
	free(res_M);
}

void Hockney_free(Hockney_model* model)
{
	if (model)
	{
		free(model->a);
		free(model->b);
		free(model);
	}
}

Hockney_model* Hockney_load(const char* filename)
{
	FILE* file = !strcmp(filename, "stdin") ? stdin : fopen(filename, "r");
	if (file == NULL)
	{
		fprintf(stderr, "Cannot open file %s\n", filename);
		return NULL;
	}
	int n;
	fscanf(file, "#Number of nodes: %d\n", &n);
	Hockney_model* model = Hockney_alloc(n);
	fscanf(file, "#p2p:\ta\t\tb\n");
	int i, index;
	for(i = 0, index = 0; i < model->n - 1; i++)
	{
		int j;
		for(j = i + 1; j < model->n; j++, index++)
			fscanf(file, "#%d-%d:\t%le\t%le\n", &n, &n, &model->a[index], &model->b[index]);
	}
	if (file != stdin)
		fclose(file);
	return model;
}

void Hockney_save(const char* filename, Hockney_model* model)
{
	FILE* file = !strcmp(filename, "stdout") ? stdout : fopen(filename, "w");
	if (file == NULL)
	{
		fprintf(stderr, "Cannot open file %s\n", filename);
		return;
	}
	fprintf(file, "#Number of nodes: %d\n", model->n);
	fprintf(file, "#p2p:\ta\t\tb\n");
	int i, index;
	for(i = 0, index = 0; i < model->n - 1; i++)
	{
		int j;
		for(j = i + 1; j < model->n; j++, index++)
			fprintf(file, "#%d-%d:\t%le\t%le\n", i, j, model->a[index], model->b[index]);
	}
	if (file != stdout)
		fclose(file);
}

double Hockney_estimate(Hockney_model* model, int i, int j, int M)
{
	int index = MPIB_IJ2INDEX(model->n, i, j); 
	return model->a[index] + model->b[index] * M;
}
