#ifndef CPM_HOCKNEY_H_
#define CPM_HOCKNEY_H_

/*!
 * \defgroup Hockney Hockney
 * \brief Computes the parameters of the Hockney model.
 * \{
 */

/*!
 * \brief Hockney model
 * \details \f$ T = \alpha + \beta M \f$
 */
typedef struct Hockney_model
{
	/*! number of nodes */
	int n;
	/*! array of \f$ C_n^2 \f$ of \f$ \alpha \f$ */
	double* a;
	/*! array of \f$ C_n^2 \f$ of \f$ \beta \f$ */
	double* b;
}
Hockney_model;

/*!
 * \brief Builds the Hockney model.  
 * \details
 * - Measures the execution time \f$ T_{ij}(0) \f$
 * of each \f$ i \xlongleftrightarrow[0]{0} j \f$ roundtrip in the communicator, \f$ i < j \f$,
 * to find \f$ \alpha = T_{ij}(0) \f$. To obtain more accurate results performs a series of roundtrips
 * and takes the average \f$ T_{ij}(0) \f$.
 * - Measures the execution time \f$ T_{ij}(M) \f$
 * of each \f$ i \xlongleftrightarrow[M]{M} j \f$ roundtrip in the communicator, \f$ i < j \f$,
 * to find \f$ \beta = \frac{T_{ij}(M) - \alpha}{M} \f$. To obtain more accurate results performs a series of roundtrips
 * and takes the average \f$ T_{ij}(M) \f$.
 * \param comm communicator, number of nodes \f$ \ge 2 \f$
 * \param M message size
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param precision measurement precision
 * \param model Hockney model (significant only at root)
 */
void Hockney_build(MPI_Comm comm, int M, int parallel, MPIB_precision precision, Hockney_model** model);

/*!
 * \brief Allocates memory for Hockney model.
 */
Hockney_model* Hockney_alloc(int n);

/*!
 * \brief Frees Hockney model.
 * \param model Hockney model
 */
void Hockney_free(Hockney_model* model);

/*!
 * \brief Loads Hockney model.
 * \param filename file name ("stdin" = stdin)
 * \return Hockney model
 */
Hockney_model* Hockney_load(const char* filename);

/*!
 * \brief Saves Hockney model.
 * \param filename file name ("stdout" = stdout)
 * \param model Hockney model
 */
void Hockney_save(const char* filename, Hockney_model* model);

/*!
 * \brief Estimates the execution time of a point-to-point communication.
 */
double Hockney_estimate(Hockney_model* model, int i, int j, int M);

/*!
 * \}
 */

#endif /*CPM_HOCKNEY_H_*/
