#ifndef CPM_MODEL_H_
#define CPM_MODEL_H_

/*!
 * \defgroup Model Model
 * \brief Computes the parameters of the communication performance model.
 * \{
 */

/*!
 * \brief CPM model
 * \details See \ref CPM_estimate_p2p, \ref CPM_estimate_one2many, \ref CPM_estimate_many2one
 */
typedef struct CPM_model
{
	/*! number of nodes */
	int n;
	/*! array of n average fixed processing delays */
	double* C;
	/*! array of \f$ C_n^2 \f$ average to/from latencies (\f$ L_{ij} = L_{ji} \f$) */
	double* L;
	/*! array of n average variable processing delays */
	double* t;
	/*! array of \f$ C_n^2 \f$ average to/from transmission rates (\f$ \beta_{ij} = \beta_{ji} \f$) */
	double* b;
	/*! threshold between small and large message sizes for the one2many model */
	int S;
	/*! one2many linear model for small message sizes */
	double one2many_small[2];
	/*! one2many linear model for large message sizes */
	double one2many_large[2];
	/*! threshold between small and medium message sizes for the many2one model */
	int M1;
	/*! threshold between medium and large message sizes for the many2one model */
	int M2;
	/*! many2one linear model for small message sizes */
	double many2one_small[2];
	/*! many2one linear model for large message sizes */
	double many2one_large[2];
}
CPM_model;

/*!
 * \brief Allocates CPM model.
 * \param n number of processors
 * \return CPM model
 */
CPM_model* CPM_model_alloc(int n);

/*!
 * \brief Frees CPM model.
 * \param model CPM model
 */
void CPM_model_free(CPM_model* model);

/*!
 * \brief Loads CPM model.
 * \param filename file name ("stdin" = stdin)
 * \return CPM model
 */
CPM_model* CPM_model_load(const char* filename);

/*!
 * \brief Saves CPM model.
 * \param filename file name ("stdout" = stdout)
 * \param model CPM model
 */
void CPM_model_save(const char* filename, CPM_model* model);

/*!
 * \brief Initializes R.
 */
int CPM_initR();

/*!
 * \brief Finalizes R.
 */
void CPM_endR();

/*!
 * \brief Builds the heterogeneous point-to-point model.  
 * \details
 * - Finds the fixed processing delays and latencies.\n
 * For each 3 nodes \f$ i < j < k \f$, measures execution times and solves systems of equations:\n
 * \f$ \begin{cases}
 * T_{ij}(0) = 2 (C_i + L_{ij} + C_j) & i \xlongleftrightarrow[0]{0} j \\
 * T_{jk}(0) = 2 (C_j + L_{jk} + C_k) & j \xlongleftrightarrow[0]{0} k \\
 * T_{ik}(0) = 2 (C_i + L_{ik} + C_k) & i \xlongleftrightarrow[0]{0} k \\
 * T_{ijk}(0) = 2 (2 C_i + \displaystyle\max_{x = j, k}(L_{ix} + C_x)) & i \xlongleftrightarrow[0]{0} jk \\
 * T_{jik}(0) = 2 (2 C_j + \displaystyle\max_{x = i, k}(L_{jx} + C_x)) & j \xlongleftrightarrow[0]{0} ik \\
 * T_{kij}(0) = 2 (2 C_k + \displaystyle\max_{x = i, j}(L_{kx} + C_x)) & k \xlongleftrightarrow[0]{0} ij \\
 * \end{cases} \f$\n
 * averages solutions:\n
 * \f$ \begin{cases}
 * C_i = (T_{ijk}(0) - \displaystyle\max_{x = j, k}T_{ix}(0)) / 2 \\
 * C_j = (T_{jik}(0) - \displaystyle\max_{x = i, k}T_{jx}(0)) / 2 \\
 * C_j = (T_{kij}(0) - \displaystyle\max_{x = i, j}T_{kx}(0)) / 2 \\
 * L_{ij} = T_{ij}(0) / 2 - (C_i + C_j) \\
 * L_{jk} = T_{jk}(0) / 2 - (C_j + C_k) \\
 * L_{ik} = T_{ik}(0) / 2 - (C_i + C_k) \\
 * \end{cases} \f$\n
 * and checks confidence intervals.
 * - Finds the variable processing delays and transmission rates.\n
 * For each 3 nodes \f$ i < j < k \f$, solves systems of equations:\n
 * \f$ \begin{cases}
 * T_{ij}(M) = 2 (C_i + L_{ij} + C_j + M (t_i + \beta_{ij} + t_j)) & i \xlongleftrightarrow[M]{M} j \\
 * T_{jk}(M) = 2 (C_j + L_{jk} + C_k + M (t_j + \beta_{jk} + t_k)) & j \xlongleftrightarrow[M]{M} k \\
 * T_{ik}(M) = 2 (C_i + L_{ik} + C_k + M (t_i + \beta_{ik} + t_k)) & i \xlongleftrightarrow[M]{M} k \\
 * T_{ijk}(M) = 2 (2 C_i + M t_i) + \displaystyle\max_{x = j, k}(2 (L_{ix} + C_x) + M (\beta_{ix} + t_x)) & i \xlongleftrightarrow[0]{M} jk \\
 * T_{jik}(M) = 2 (2 C_j + M t_j) + \displaystyle\max_{x = i, k}(2 (L_{jx} + C_x) + M (\beta_{jx} + t_x)) & j \xlongleftrightarrow[0]{M} ik \\
 * T_{kij}(M) = 2 (2 C_k + M t_k) + \displaystyle\max_{x = i, j}(2 (L_{kx} + C_x) + M (\beta_{kx} + t_x)) & k \xlongleftrightarrow[0]{M} ij \\
 * \end{cases} \f$\n
 * averages solutions:\n
 * \f$ \begin{cases}
 * t_i = (T_{ijk}(M) - \displaystyle\max_{x = j, k}(T_{ix}(0) + T_{ix}(M)) / 2 - 2 C_i) / M \\
 * t_j = (T_{jik}(M) - \displaystyle\max_{x = i, k}(T_{jx}(0) + T_{jx}(M)) / 2 - 2 C_j) / M \\
 * t_j = (T_{kij}(M) - \displaystyle\max_{x = i, j}(T_{kx}(0) + T_{kx}(M)) / 2 - 2 C_k) / M \\
 * \beta_{ij} = (T_{ij}(M) / 2 - (C_i + L_{ij} + C_j)) / M - (t_i + t_j) \\
 * \beta_{jk} = (T_{jk}(M) / 2 - (C_j + L_{jk} + C_k)) / M - (t_j + t_k) \\
 * \beta_{ik} = (T_{ik}(M) / 2 - (C_i + L_{ik} + C_k)) / M - (t_i + t_k) \\
 * \end{cases} \f$\n
 * and checks confidence intervals.
 * 
 * It is called by \ref CPM_build.
 * \param model CPM model, must be allocated and filled by \ref CPM_build_one2many and \ref CPM_build_many2one (significant only at root)
 * \param comm communicator, number of nodes \f$ \ge 3 \f$
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param precision measurement parameters
 */
void CPM_build_p2p(CPM_model* model, MPI_Comm comm, int parallel, MPIB_precision precision);

void CPM_build_p2p_(CPM_model* model, MPI_Comm comm, int parallel, MPIB_precision precision);

/*!
 * \brief Computes the parameters of one-to-many.
 * \details Measures one-to-many execution time for the message sizes \f$ 0 < M < max\_size \f$ and
 * performs the Bai & Perron algorithm over the F statistic for the data to find the \f$ S \f$ breakpoint
 * in the piecewise linear regression \f$ T \f$ ~ \f$ M \f$. R must be initialized by \ref CPM_initR at root beforehand.
 * 
 * It is called by \ref CPM_build.
 * \param model CPM model, must be allocated (significant only at root)
 * \param comm communicator
 * \param stride stride for message sizes
 * \param max_size maximum message size
 * \param precision measurement parameters
 */
void CPM_build_one2many(CPM_model* model, MPI_Comm comm, int stride, int max_size, MPIB_precision precision);

/*!
 * \brief Computes the parameters of many-to-one.
 * \details Measures many-to-one execution time for the message sizes \f$ 0 < M < max\_size \f$ and
 * performs the Bai & Perron algorithm over the F statistic for the data to find the \f$ M_2 \f$ breakpoint
 * in the piecewise linear regression \f$ T \f$ ~ 1. R must be initialized by \ref CPM_initR at root beforehand.
 *
 * Then in the loop measures many-to-one execution time for the message sizes \f$ 0 < M < m \f$
 * with the stride reduced twice each time. \f$ m \f$ is a message size for which tenfold escalation of the execution time
 * has been observed on the previous step. As stride reaches 1-byte value \f$ m \f$ is truncated to a kb value.
 * 
 * It is called by \ref CPM_build.
 * \param model CPM model, must be allocated (significant only at root)
 * \param comm communicator
 * \param stride stride for message sizes
 * \param max_size maximum message size
 * \param precision measurement parameters
 */
void CPM_build_many2one(CPM_model* model, MPI_Comm comm, int stride, int max_size, MPIB_precision precision);

/*!
 * \brief Computes the parameters of CPM model.
 * \details Calls \ref CPM_build_one2many, \ref CPM_build_many2one, \ref CPM_build_p2p.
 * R must be initialized by \ref CPM_initR at root beforehand.
 * \param comm communicator
 * \param stride stride for message sizes
 * \param max_size maximum message size
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param precision measurement parameters
 * \param model CPM model (significant only at root)
 */
void CPM_build(MPI_Comm comm, int stride, int max_size, int parallel, MPIB_precision precision, CPM_model** model);

/*!
 * \brief Estimates the execution time of point-to-point communication.
 * \details The execution time of \f$ i \xlongrightarrow{M} j \f$ is equal to\n
 * \f$ C_i + L_{ij} + C_j + M (t_i + \beta_{ij} + t_j)\f$ 
 * \param model CPM model
 * \param i index of the precessor
 * \param j index of the precessor
 * \param M message size
 * \return predicted execution time
 */
double CPM_estimate_p2p(CPM_model* model, int i, int j, int M);

/*!
 * \brief Estimates the execution time of one-to-many communication.
 * \details The execution time of \f$ 0 \xlongrightarrow{M} 1..n - 1 \f$ is equal to\n
 * \f$ (n - 1) (C_0 + M t_i) + \begin{cases}
 * \displaystyle\max_{i = 1}^{n - 1} (L_{0i} + C_i + M (\beta_{0i} + t_i)) & M < S \\
 * \displaystyle\sum_{i = 1}^{n - 1} (L_{0i} + C_i + M (\beta_{0i} + t_i)) & M \ge S \\
 * \end{cases} \f$ 
 * \param model CPM model
 * \param root root precessor
 * \param M message size
 * \return predicted execution time
 */
double CPM_estimate_one2many(CPM_model* model, int root, int M);

/*!
 * \brief Estimates the execution time of many-to-one communication.
 * \details The execution time of \f$ 0 \xlongrightarrow{M} 1..n-1 \f$ is equal to\n
 * \f$ (n - 1) (C_0 + M t_i) + \begin{cases}
 * \displaystyle\max_{i = 1}^{n - 1} (L_{0i} + C_i + M (\beta_{0i} + t_i)) & M < M_1 \\
 * \displaystyle\sum_{i = 1}^{n - 1} (L_{0i} + C_i + M (\beta_{0i} + t_i)) & M > M_2 \\
 * \end{cases} \f$ 
 * \param model CPM model
 * \param root root precessor
 * \param M message size
 * \return predicted execution time
 */
double CPM_estimate_many2one(CPM_model* model, int root, int M);

/*!
 * \}
 */

#endif /*CPM_MODEL_H_*/
