#include "cpm.h"
#ifdef HAVE_LIBLOGP

#include <logp_mpi.h>
#include <malloc.h> 
#include <memory.h> 
#include <string.h>
#include <stdio.h>

/*!
 * \brief Allocates memory for PLogP model.
 */
PLogP_model* PLogP_alloc(int n)
{
	PLogP_model* model = malloc(sizeof(PLogP_model));
	model->n = n;
	model->logp = malloc(sizeof(logp_params*) * n * (n - 1) / 2);
	return model;
}

/*!
 * \brief Gets the size of logp data structure
 */
int PLogP_size(logp_params* logp)
{
	int size = sizeof(int);// nsizes
	size += sizeof(double);// L
	size += sizeof(logp_size_params) * logp->nsizes;// params
	return size;
}

/*!
 * \brief Packs PLogP model to send via MPI
 */
void* pack(logp_params* logp)
{
	void* buffer = malloc(PLogP_size(logp));
	void* pos = buffer;
	memcpy(pos, &logp->nsizes, sizeof(int));
	pos += sizeof(int);
	memcpy(pos, &logp->L, sizeof(double));
	pos += sizeof(double);
	memcpy(pos, logp->params, sizeof(logp_size_params) * logp->nsizes);
	return buffer;
}

/*!
 * \brief Unpacks PLogP model after receive via MPI
 */
logp_params* unpack (void* buffer)
{
	logp_params* logp = malloc(sizeof(logp_params));
	logp->descr = malloc(sizeof(char) * 4);
	strcpy(logp->descr, "CPM");
	void* pos = buffer;
	memcpy(&logp->nsizes, pos, sizeof(int));
	memcpy(&logp->nsizes_max, pos, sizeof(int));
	pos += sizeof(int);
	memcpy(&logp->L, pos, sizeof(double));
	pos += sizeof(double);
	int size = sizeof(logp_size_params) * logp->nsizes;
	logp->params = malloc(size);
	memcpy(logp->params, pos, size);
	return logp;
}

void PLogP_build(MPI_Comm comm, int parallel, MPIB_precision precision, PLogP_model** model_)
{
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	if (size < 2)
	{
		if (rank == 0)
			fprintf(stderr, "Cannot compute logp parameters for %d processors (must be >= 2)\n", size);
		return;
	}

	int sendcount = size - rank - 1;
	logp_params** logp = malloc(sizeof(logp_params*) * sendcount);

	MPI_Group group;
	MPI_Comm_group(comm, &group);
	int index = 0;
	MPIB_pairs* pairs = MPIB_build_pairs(size);
	MPIB_pairs* iter_pairs = pairs;
	while (iter_pairs)
	{
		MPI_Barrier(comm);
		MPIB_pair* iter_pair = iter_pairs->list;
		while (iter_pair)
		{
			if (!parallel)
				MPI_Barrier(comm);
			MPI_Group group_pair;
			MPI_Group_incl(group, 2, iter_pair->values, &group_pair);
			MPI_Comm comm_pair;
			MPI_Comm_create(comm, group_pair, &comm_pair);
			if (comm_pair != MPI_COMM_NULL)
			{
				logp_params* tmp;
				logp_measure(&tmp, "CPM", comm_pair, 0, 1, 0,
					precision.min_reps, precision.max_reps, precision.cl, precision.eps,
					LOGP_DEF_MAX_DIFF, LOGP_DEF_MIN_SIZE, LOGP_DEF_MAX_SIZE, LOGP_DEF_DIFF_SIZES,
					LOGP_MPI_SEND, LOGP_MPI_RECV, LOGP_BUF_DYNAMIC, LOGP_OPT_GM_DIRECT, LOGP_DIAG_SILENT);
				if (rank == iter_pair->values[0])
					logp[index++] = tmp;
				MPI_Comm_free(&comm_pair);
			}
			MPI_Group_free(&group_pair);
			iter_pair = iter_pair->next;
		}
		iter_pairs = iter_pairs->next;
	}
	MPIB_free_pairs(pairs);
	MPI_Group_free(&group);

	if (rank == 0)
	{
		PLogP_model* model = PLogP_alloc(size);
		int i, j, index;
		for (i = 0, index = 0; i < sendcount; i++, index++)
			model->logp[i] = logp[i];
		for (i = 1; i < size - 1; i++)
		{
			for (j = i + 1; j < size; j++, index++)
			{
				int size;
				MPI_Recv(&size, 1, MPI_INT, i, 0, comm, MPI_STATUS_IGNORE);
				void* buffer = malloc(size);
				MPI_Recv(buffer, size, MPI_CHAR, i, 0, comm, MPI_STATUS_IGNORE);
				model->logp[index] = unpack(buffer);
				free(buffer);
			}
		}
		*model_ = model;
	}
	else
	{
		int i;
		for (i = 0; i < sendcount; i++)
		{
			int size = PLogP_size(logp[i]);
			MPI_Send(&size, 1, MPI_INT, 0, 0, comm);
			void* buffer = pack(logp[i]);
			MPI_Send(buffer, size, MPI_CHAR, 0, 0, comm);
			free(buffer);
			logp_free(logp[i]);
		}
	}
	free(logp);
}

void PLogP_free(PLogP_model* model)
{
	if (model)
	{
		int i;
		int size;
		for (i = 0, size = model->n * (model->n - 1) / 2; i < size; i++)
			logp_free(model->logp[i]);
		free(model);
	}
}

PLogP_model* PLogP_load(const char* filename)
{
	FILE* file = !strcmp(filename, "stdin") ? stdin : fopen(filename, "r");
	if (file == NULL)
	{
		fprintf(stderr, "Cannot open file %s\n", filename);
		return NULL;
	}
	int n;
	fscanf(file, "#Number of nodes: %d\n", &n);
	PLogP_model* model = PLogP_alloc(n);

	char* tmpfilename;
	if (filename == NULL)
	{
		tmpfilename = malloc(sizeof(char) * 6);
		strcpy(tmpfilename, "plogp");
	}
	else
	{
		tmpfilename = malloc(sizeof(char) * (strlen(filename) + 6));
		strcpy(tmpfilename, filename);
		strcat(tmpfilename, ".plogp");
	}
	int i, index;
	for (i = 0, index = 0; i < model->n - 1; i++)
	{
		int j;
		for (j = i + 1; j < model->n; j++, index++)
		{
			int tmp;
			fscanf(file, "#%d-%d\n", &tmp, &tmp);
			FILE* tmpfile = fopen(tmpfilename, "w");
			if (tmpfile != NULL)
			{
				char c;
				while((c = fgetc(file)) != EOF)
				{
					fputc(c, tmpfile);
					if (c == '\n')
					{
						if ((c = fgetc(file)) == '\n')
							break;
						else
							fputc(c, tmpfile);
					}
				}
				fclose(tmpfile); 
			}
			logp_load(model->logp[index], tmpfilename);
		}
	}
	free(tmpfilename);

	if (file != stdin)
		fclose(file);
	return model;
}

void PLogP_save(const char* filename, PLogP_model* model)
{
	FILE* file = !strcmp(filename, "stdout") ? stdout : fopen(filename, "w");
	if (file == NULL)
	{
		fprintf(stderr, "Cannot open file %s\n", filename);
		return;
	}
	fprintf(file, "#Number of nodes: %d\n", model->n);

	char* tmpfilename;
	if (filename == NULL)
	{
		tmpfilename = malloc(sizeof(char) * 6);
		strcpy(tmpfilename, "plogp");
	}
	else
	{
		tmpfilename = malloc(sizeof(char) * (strlen(filename) + 6));
		strcpy(tmpfilename, filename);
		strcat(tmpfilename, ".plogp");
	}
	int i, index;
	for (i = 0, index = 0; i < model->n - 1; i++)
	{
		int j;
		for (j = i + 1; j < model->n; j++, index++)
		{
			fprintf(file, "#%d-%d\n", i, j);
			logp_save(model->logp[index], tmpfilename);
			FILE* tmpfile = fopen(tmpfilename, "r");
			if (tmpfile != NULL)
			{
				char c;
				while((c = fgetc(tmpfile)) != EOF)
					fputc(c, file);
				fclose(tmpfile); 
			}
			fprintf(file, "\n\n");
		}
	}
	free(tmpfilename);

	if (file != stdout)
		fclose(file);
}

double PLogP_estimate(PLogP_model* model, int i, int j, int M)
{
	logp_params* logp = (logp_params*)model->logp[MPIB_IJ2INDEX(model->n, i, j)];
	return 2.0 * logp_L(logp) + logp_g(logp, 0) + logp_g(logp, M);
}

double PLogP_estimate_LogGP(PLogP_model* model, int i, int j, int M)
{
	logp_params* logp = (logp_params*)model->logp[MPIB_IJ2INDEX(model->n, i, j)];
	double T = logp_L(logp) + logp_g(logp, 1);
	if (M > 0)
		T += logp_g(logp, logp->params[logp->nsizes - 1].size) * (M - 1) / logp->params[logp->nsizes - 1].size;
	return T;
}

#endif /*HAVE_LIBLOGP*/
