#ifndef CPM_PLOGP_H_
#define CPM_PLOGP_H_

/*!
 * \defgroup PLogP PLogP
 * \brief Computes the parameters of the parameterized LogP model.
 * \details PLogP is an extension of the LogP model that describes a network in terms of the following parameters:
 * - L : latency - this is the end-to-end latency between nodes 
 * - o : overhead  - for sending and receiving \f$ o_{s}(m)\f$ and \f$ o_{r}(m) \f$ respectively for message size \f$ m \f$
 * - g : gap per message   depends on message size \f$ g(m) \f$ 
 * - P : number of nodes involved in communication 
 * 
 * The PLogP model is defined in terms of these parameters,  the end-to-end latency L, 
 * sender and receiver overheads, \f$o_{s}(m) \f$ and \f$o_{r}(m)\f$ respectively, gap per message g(m),
 * and number of nodes involved in communication \f$ P \f$. In this model sender and receiver overheads
 * and gap per message depend on the message size.
 * Time to send a message of size \f$ m \f$ between two nodes in the PLogP model is \f$ L + g(m)\f$.
 * \{
 */

/*!
 * \brief PLogP model
 * \details See \ref PLogP_estimate.
 */
typedef struct PLogP_model
{
	/*! number of nodes */
	int n;
	/*! array of \f$ C_n^2 \f$ p2p precision */
	void** logp;
}
PLogP_model;

/*!
 * \brief Builds PLogP model.
 * \details Calls logp_mpi library.
 * \param comm communicator
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param precision measurement precision
 * \param model PLogP model
 */
void PLogP_build(MPI_Comm comm, int parallel, MPIB_precision precision, PLogP_model** model);

/*!
 * \brief Frees PLogP model.
 */
void PLogP_free(PLogP_model* model);

/*!
 * \brief Loads PLogP model.
 * \param filename file name ("stdin" = stdin)
 * \return PLogP model
 */
PLogP_model* PLogP_load(const char* filename);

/*!
 * \brief Saves PLogP model.
 * \param filename file name ("stdout" = stdout)
 * \param model PLogP model
 */
void PLogP_save(const char* filename, PLogP_model* model);

/*!
 * \brief Estimates the execution time of a point-to-point communication.
 */
double PLogP_estimate(PLogP_model* model, int i, int j, int M);

/*!
 * \brief Estimates the execution time of a point-to-point communication according to LogGP model.
 * \details \f$ T = L + 2 * o + G (M - 1) \f$\n
 * The meaning of the parameters of LogGP and PLogP models:\n
 * - \f$ L = L_p + g_p(1) - os_p(1) - or_p(1) \f$
 * - \f$ 2 * o = os_p(1) + or_p(1) \f$
 * - \f$ G = g_p(M_{max}) / M_{max} \f$
 * 
 * \param model PLogP model
 * \param i index of the process
 * \param j index of the process
 * \param M message size
 */
double PLogP_estimate_LogGP(PLogP_model* model, int i, int j, int M);

/*!
 * \}
 */

#endif /*CPM_PLOGP_H_*/
